/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.chemistry;

import com.ggasoftware.indigo.Indigo;
import com.ggasoftware.indigo.IndigoObject;
import com.google.common.base.Preconditions;
import java.text.ParseException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import lombok.AllArgsConstructor;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.bitbucket.kienerj.moleculedatabaseframework.io.PropertySetter;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfRecord;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;

/**
 * Generates ChemicalCompound instances from the given input data.
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@AllArgsConstructor
public class ChemicalCompoundFactory<T extends ChemicalCompound> {

    private static final XLogger logger = XLoggerFactory.getXLogger("ChemicalCompoundFactory");    
    private final ChemicalStructureFactory chemicalStructureFactory;
    private final Class<T> compoundClass;

    /**
     * <p> Creates a
     * <code>ChemicalCompound</code> from an
     * <code>SdfRecord</code> with compositions and all mapped properties set.
     * </p> <p> See
     * {@link org.bitbucket.kienerj.moleculedatabaseframework.service.ChemicalCompoundService#importSDF(Reader,Map,Map,String,boolean) importSDF}
     * for further information especially on property mapping. </p> <p> In case
     * compoundName is not mapped, compoundName is set to molfileName. </p>
     *
     * @param sdfRecord
     * @param percentages
     * @param mappedProperties
     * @return
     * @throws ParseException when the given data can not be converted to the
     * target type
     * @throws IllegalAccessException
     * @throws NoSuchFieldException in case of an invalid mapping      * in <code>mappedProperties</code>
     * @throws InstantiationException if <code>compoundClass</code> does not
     * have a zero argument constructor
     */
    public T createChemicalCompound(SdfRecord sdfRecord, Map<String, Double> percentages,
            Map<String, String> mappedProperties) throws ParseException,
            IllegalAccessException, NoSuchFieldException, InstantiationException {

        logger.entry(sdfRecord, percentages, mappedProperties);
        logger.debug("Creating ChemicalCompound {}.", sdfRecord.getMolfileName());
        Preconditions.checkNotNull(sdfRecord);

        String molfile = sdfRecord.getMolfile();
        Map<String, String> properties = sdfRecord.getProperties();
        T compound = createChemicalCompoundFromMolfile(molfile, percentages);
        if (mappedProperties == null) {
            mappedProperties = new HashMap<>();
        }
        if (!mappedProperties.containsValue("compoundName")) {
            logger.info("No compoundName propery found. Trying to assign compoundName from molfile name.");
            mappedProperties.put("compoundName", "compoundName");
            properties.put("compoundName", sdfRecord.getMolfileName());
        }

        PropertySetter.setProperties(compound, compoundClass, mappedProperties, properties);
        logger.exit(compound);
        return compound;
    }

    /**
     * <p> Create a
     * <code>ChemicalCompound</code> from a molfile. If the molfile contains
     * disconnected structures (is a mixture), you must supply the relative
     * amount of each structure. </p> <p> Returns a
     * <code>ChemicalCompound</code> with compositions set. All other properties
     * must be set outside this method. </p>
     *
     * @param molfile
     * @param percentages key = InChIkey, value = percentage of the structure in
     * the ChemicalCompound
     * @return a ChemicalCompound with compositions properly set.
     * @throws IllegalAccessException
     * @throws InstantiationException
     */
    private T createChemicalCompoundFromMolfile(String molfile, Map<String, Double> percentages) throws
            IllegalAccessException, InstantiationException {

        logger.entry(molfile, percentages);

        Indigo indigo = new Indigo();
        IndigoObject mol = indigo.loadMolecule(molfile);

        Set<ChemicalCompoundComposition> compositions = new HashSet<>();

        T compound = compoundClass.newInstance();

        for (IndigoObject component : mol.iterateComponents()) {
            
            IndigoObject clone;

            if (mol.countComponents() > 1) {
                clone = component.clone();
            } else {
                // in case of single structure in molfile
                // store the original molfile as uploaded
                clone = mol;
            }

            ChemicalStructure structure = chemicalStructureFactory
                    .createChemicalStructure(clone);

            Double percentage = null;
            if (percentages != null) {
                percentage = percentages.get(
                        structure.getStructureKey());
                logger.debug("Assigning ChemicalCompoundCompostion Percentage.");
            }
            if (percentage == null
                    && mol.countComponents() == 1) {
                logger.debug("No percentage found. Assigning default of 100 for \"single composition\" ChemicalCompound.");
                percentage = 100.0;
            }
            if (structure != null) {
                ChemicalCompoundComposition composition = new ChemicalCompoundComposition(
                        compound, structure, percentage);
                compositions.add(composition);
            }
        }

        compound.setCompositions(compositions);
        logger.exit(compound);
        return compound;
    }
}
