/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.chemistry;

import com.ggasoftware.indigo.*;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class ChemicalFormatConverter {
    /*
     * Indigo state maschine is generated separatley in each method call because
     * it should not be shared across threads.
     */

    private static final XLogger logger = XLoggerFactory.getXLogger("ChemicalFormatConverter");

    /**
     * <p>Converts the given molfile or SMILES
     * <code>structureData</code> to SMILES</p> <p>The returned smiles follows
     * the smiles spec and has no extensions appended.</p>
     *
     * @param structureData the molfile
     * @return the smiles string
     */
    public String toSmiles(String structureData) {

        return toSmiles(structureData, true);
    }

    /**
     * <p>Converts the given molfile or SMILES
     * <code>structureData</code> to SMILES</p> <p>If
     * <code>removeSmilesExtensions</code> is true, all extensions are removed
     * (everything after the first space in the string). </p> <p> See
     * http://www.chemaxon.com/marvin/help/formats/cxsmiles-doc.html for some of
     * the possible extensions used by indigo. </p>
     *
     * @param structureData the molfile
     * @return the smiles string
     */
    public String toSmiles(String structureData, boolean removeSmilesExtensions) {

        logger.entry(structureData, removeSmilesExtensions);
        try {
            Indigo indigo = new Indigo();
            IndigoObject mol = indigo.loadMolecule(structureData);
            mol.aromatize();
            mol.layout();
            String smiles = mol.canonicalSmiles();
            /*
             * Indigo may add SMILES extensions after the smiles string separated by
             * a space.
             * As example molfiles with stereo bonds but without chiral flag
             * get converted to smiles that have additional stereo info at end of
             * string:
             *
             * "CC[C@@H](N)[C@@H](O)CC |&1:2,4|"
             *
             * see http://www.chemaxon.com/marvin/help/formats/cxsmiles-doc.html
             *      - AND stereo group type:
             *          &<group>:<atomindex>,< atomindex>...
             *
             * This happens for non-chiral molecules with up/down bonds.
             *
             * if chiral flag is set, absolute stereo chemistry is assumed
             * and no SMILES extension is set.
             * Chiral flag is on molfile line 4, the 5th option (the 1):
             *      8  8  0  0  1  0  0  0  0  0999 V2000
             *
             * Below we optionally remove the extensions as they do not belong to
             * the smiles specification. In above case they are from chemaxon.
             */
            if (removeSmilesExtensions) {
                int firstSpaceIndex = smiles.indexOf(" ");
                if (firstSpaceIndex != -1) {
                    logger.warn("The input molfile did not have chiral flag set.");
                    smiles = smiles.substring(0, firstSpaceIndex);
                }
            }
            logger.exit(smiles);
            return smiles;
        } catch (IndigoException ex) {
            throw new FormatConversionException(ex, structureData);
        }
    }

    /**
     * <p>Converts the given SMILES String or molfile
     * <code>structureData</code> to a molfile with aromatic bond type set.</p>
     *
     * @param structureData smiles string
     * @return the molfile
     */
    public String toMol(String structureData) {
        return toMol(structureData, false);
    }

    /**
     * <p>Converts the given SMILES String or molfile
     * <code>structureData</code> to a molfile.</p>
     * <p>
     * if <code>aromatize</code> is true this will set bond order to for aromatic bonds to 4.
     * </p>
     *
     * @param structureData smiles string
     * @param aromatize whether to use aromatic bonds or single-double bonds
     * @return the molfile
     */
    public String toMol(String structureData, boolean aromatize) {

        logger.entry(structureData);
        try {
            Indigo indigo = new Indigo();
            IndigoObject mol = indigo.loadMolecule(structureData);
            if (aromatize) {
                mol.aromatize();
            } else {
                mol.dearomatize();
            }
            mol.layout(); // absolutley required else you loose stereo chemistry from SMILES!!!
            String molFile = mol.molfile();
            logger.exit(molFile);
            return molFile;
        } catch (IndigoException ex) {
            throw new FormatConversionException(ex, structureData);
        }
    }

    /**
     * <p>Converts the given SMILES or molfile in
     * <code>structureData</code> to a standard InChi.</p>
     *
     * @param structureData molfile or SMILES String
     * @return the standard InChi
     */
    public String toStandardInchi(String structureData) {

        logger.entry(structureData);
        try {
            Indigo indigo = new Indigo();
            IndigoInchi indigoInchi = new IndigoInchi(indigo);
            IndigoObject mol = indigo.loadMolecule(structureData);
            String inchi = indigoInchi.getInchi(mol);
            logger.exit(inchi);
            return inchi;
        } catch (IndigoException ex) {
            throw new FormatConversionException(ex, structureData);
        }
    }

    /**
     * <p>Converts the given SMILES or molfile in
     * <code>structureData</code> to a standard InChikey.</p>
     *
     * @param structureData molfile or SMILES String
     * @return the standard InChikey
     */
    public String toStandardInchikey(String structureData) {

        logger.entry(structureData);
        try {
            Indigo indigo = new Indigo();
            IndigoInchi indigoInchi = new IndigoInchi(indigo);
            IndigoObject mol = indigo.loadMolecule(structureData);
            mol.unfoldHydrogens();
            String inchi = indigoInchi.getInchi(mol);
            String inchiKey = indigoInchi.getInchiKey(inchi);
            logger.exit(inchiKey);
            return inchiKey;
        } catch (IndigoException ex) {
            throw new FormatConversionException(ex, structureData);
        }
    }

    /**
     * <p>Converts the given SMILES or molfile in
     * <code>structureData</code> to a InChi using the given options.</p>
     *
     * <p>Example options Strings:<br> "/DoNotAddH /SUU /SLUUD"<br> "-DoNotAddH
     * /SUU -SLUUD"<br> </p>
     *
     * @param structureData molfile or SMILES String
     * @param inchiOptions options to generate a non-standard InChi
     * @return the standard InChi
     */
    public String toInchi(String structureData, String inchiOptions) {

        logger.entry(structureData, inchiOptions);
        try {
            Indigo indigo = new Indigo();
            indigo.setOption("inchi-options", inchiOptions);
            IndigoInchi indigoInchi = new IndigoInchi(indigo);
            IndigoObject mol = indigo.loadMolecule(structureData);
            String inchi = indigoInchi.getInchi(mol);
            logger.exit(inchi);
            return inchi;
        } catch (IndigoException ex) {
            throw new FormatConversionException(ex, structureData);
        }
    }

    /**
     * <p>Converts the given InChi
     * <code>inchi</code> to its InChikey.</p>
     *
     * <p>If the InChi is not standard, the key will also not be a standard
     * InChikey.</p>
     *
     * @param inchi the InChi to convert to an InChikey
     * @return the standard InChikey
     */
    public String inchiToInchikey(String inchi) {

        logger.entry(inchi);
        try {
            Indigo indigo = new Indigo();
            IndigoInchi indigoInchi = new IndigoInchi(indigo);
            String inchiKey = indigoInchi.getInchiKey(inchi);
            logger.exit(inchiKey);
            return inchiKey;
        } catch (IndigoException ex) {
            throw new FormatConversionException(ex, inchi);
        }
    }

    /**
     * <p>Converts the given InChi
     * <code>inchi</code> to a SMILES string.</p>
     *
     * @param inchi
     * @return
     */
    public String inchiToSmiles(String inchi) {

        logger.entry(inchi);
        try {
            Indigo indigo = new Indigo();
            IndigoInchi indigoInchi = new IndigoInchi(indigo);
            IndigoObject mol = indigoInchi.loadMolecule(inchi);
            mol.aromatize();
            String result = mol.canonicalSmiles();
            logger.exit(result);
            return result;
        } catch (IndigoException ex) {
            throw new FormatConversionException(ex, inchi);
        }
    }

    /**
     * <p>Converts the given InChi
     * <code>inchi</code> to a molfile.</p> <p>Explicit hydrogens are
     * removed.</p>
     *
     * @param inchi
     * @return
     */
    public String inchiToMol(String inchi) {

        logger.entry(inchi);
        try {
            Indigo indigo = new Indigo();
            IndigoInchi indigoInchi = new IndigoInchi(indigo);
            IndigoObject mol = indigoInchi.loadMolecule(inchi);
            mol.foldHydrogens();
            mol.aromatize();
            mol.layout();
            String result = mol.molfile();
            logger.exit(result);
            return result;
        } catch (IndigoException ex) {
            throw new FormatConversionException(ex, inchi);
        }
    }
}
