/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.entity;

import com.fasterxml.jackson.annotation.JsonIdentityInfo;
import com.fasterxml.jackson.annotation.ObjectIdGenerators;
import com.mysema.query.annotations.QueryInit;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.DiscriminatorColumn;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.SequenceGenerator;
import javax.persistence.Table;
import lombok.Data;
import lombok.EqualsAndHashCode;

/**
 * <p>
 * Superclass for all classes containing exactly 1 <code>Containable</code>.
 * </p>
 * <p>
 * There should only be 1 implementation of ChemicalCompoundContainer in the application.
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Entity
@Inheritance(strategy = InheritanceType.SINGLE_TABLE)
@Table(name="chemical_compound_container")
@DiscriminatorColumn(name="container_type")
@JsonIdentityInfo(generator=ObjectIdGenerators.IntSequenceGenerator.class, property="@id")
@Data
@EqualsAndHashCode(callSuper=false, of = {"barcode"})
public abstract class ChemicalCompoundContainer<T extends Containable>
        extends BaseEntity {

    @Id
    @GeneratedValue(strategy = GenerationType.AUTO, generator = "compound_container_seq_gen")
    @SequenceGenerator(name = "compound_container_seq_gen",
            sequenceName = "seq_compound_container", allocationSize = 1000)
    @Column(name = "compound_container_id")
    private Long id;

    @Column(name = "barcode", unique = true, nullable = false)
    private String barcode;

    @QueryInit({"chemicalCompound.readRole.users", "readRole.users", "id"})
    @JoinColumn(name = "containable_id", nullable = false)
    @ManyToOne(fetch = FetchType.EAGER, cascade = {CascadeType.PERSIST, CascadeType.REFRESH},
            targetEntity=Containable.class)
    private T containable;

    /**
     * For JPA Provider
     */
    protected ChemicalCompoundContainer(){}

    public ChemicalCompoundContainer(Long id, String barcode, T containable){
        this.id = id;
        this.containable = containable;
        this.barcode = barcode;
    }

    public ChemicalCompoundContainer( String barcode, T containable){
        this(null, barcode, containable);
    }

    public ChemicalCompound<T> getChemicalCompound(){
        return containable.getChemicalCompound();
    }

    public void setContainable(Containable containable) {
        this.containable = (T)containable;
    }
}