/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.entity;

import com.fasterxml.jackson.annotation.JsonIdentityInfo;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.ObjectIdGenerators;
import java.util.HashSet;
import java.util.Set;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.SequenceGenerator;
import javax.persistence.Table;
import lombok.Data;
import lombok.EqualsAndHashCode;

/**
 * <p>
 * Linker Class between Compound and Container.
 * </p>
 * <p>
 * Consider you create Compound A in your factory. Each Production run (Lot) has
 * slightly different properties. What you actually sell to a customer is a
 * container containing Lot 1323 of Compound A and not just Compound A directly.
 * </p>
 * <p>
 * Implementations should have 1 unique field clearly identifying the record like
 * lotNumber in above example (id does not count).
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Entity
@Inheritance(strategy = InheritanceType.TABLE_PER_CLASS)
@Table(name="containable")
@JsonIdentityInfo(generator=ObjectIdGenerators.IntSequenceGenerator.class, property="@id")
@Data
@EqualsAndHashCode(callSuper = false, of = {"chemicalCompound"})
public abstract class Containable<T extends ChemicalCompound> extends BaseEntity {

    @Id
    @GeneratedValue(strategy = GenerationType.AUTO, generator = "containable_seq_gen")
    @SequenceGenerator(name = "containable_seq_gen",
    sequenceName = "seq_containable", allocationSize = 1000)
    @Column(name = "id")
    private Long id;

    @JoinColumn(name = "chemical_compound_id", nullable = false)
    @ManyToOne(fetch = FetchType.EAGER, cascade = {CascadeType.PERSIST, CascadeType.REFRESH},
            targetEntity=ChemicalCompound.class)
    private T chemicalCompound;

    @OneToMany(fetch = FetchType.LAZY, mappedBy = "containable",
        targetEntity=ChemicalCompoundContainer.class, cascade = {CascadeType.PERSIST,
            CascadeType.REFRESH},
        orphanRemoval = true)
    private Set<ChemicalCompoundContainer> containers = new HashSet<>();

    @JsonIgnore
    @OneToOne(fetch = FetchType.LAZY)
    private Role readRole;

    public Containable() {
    }

    public Containable(Long id, T chemicalCompound) {
        this.id = id;
        this.chemicalCompound = chemicalCompound;
    }

    public Containable(T chemicalCompound) {
        this(null, chemicalCompound);
    }

    public void addChemicalCompoundContainer(ChemicalCompoundContainer container){
        containers.add(container);
    }
}
