/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.exceptionhandler;

import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.annotation.AfterThrowing;
import org.aspectj.lang.annotation.Aspect;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.BaseEntity;
import org.bitbucket.kienerj.moleculedatabaseframework.service.Service;
import org.bitbucket.kienerj.moleculedatabaseframework.sql.EntityVersionConflictException;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.core.annotation.Order;
import org.springframework.orm.ObjectOptimisticLockingFailureException;
import org.springframework.stereotype.Component;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Component
@Order(3000) // must be higher than order of <tx:annotation-driven />
@Aspect
public class OLExceptionHandler {

    private static final XLogger logger = XLoggerFactory.getXLogger("OLExceptionHandler");

    @AfterThrowing(pointcut = "execution(* org.bitbucket.kienerj.moleculedatabaseframework.service.Service.save(*))", throwing = "ex")
    public void handleException(JoinPoint joinPoint, ObjectOptimisticLockingFailureException ex) {

        logger.entry(joinPoint, ex);
        Service service = (Service) joinPoint.getTarget();
        // save() method always only has 1 argument
        BaseEntity conflictingVersion = (BaseEntity) joinPoint.getArgs()[0];

        // should only happen when updating existing entity (eg. merging)
        // and because entites do not use CascadeType.MERGE
        // the entity causing the issue should always be of class
        // this.entityClass
        // NOTE: for some reason lockEx.getPersistentClass() returns null!!!
        // hence comparing by class name...
        if (ex.getPersistentClassName().equals(service.getEntityClass().getName())) {
            BaseEntity currentVersion = (BaseEntity)service.getById(conflictingVersion.getId());
            throw new EntityVersionConflictException(ex, currentVersion, conflictingVersion);
        } else {
            throw ex;
        }
    }
}
