/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import com.ggasoftware.indigo.*;
import com.google.common.base.Preconditions;
import java.io.IOException;
import java.io.OutputStream;
import lombok.AllArgsConstructor;
import org.bitbucket.kienerj.moleculedatabaseframework.MoleculeDatabaseFrameworkException;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@AllArgsConstructor
public class IndigoMoleculeRenderer {

    private int width;
    private int height;

    /**
     * Renders image to Outputstream
     *
     *
     * @param mol
     * @param out
     */
    public void renderChemicalStructure(String structureData, OutputStream out) {

        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(out);

        Indigo indigo = new Indigo();
        IndigoRenderer renderer = new IndigoRenderer(indigo);
        setRenderingOptions(indigo);

        IndigoObject mol = indigo.loadMolecule(structureData);
        byte[] buffer = renderer.renderToBuffer(mol);
        try {
            out.write(buffer);
        } catch (IOException ioEx) {
            throw new MoleculeDatabaseFrameworkException(ioEx);
        }
    }

    public void renderChemicalStructure(String structureData, OutputStream out,
            String highlightedSubstructure) {

        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(out);

        Indigo indigo = new Indigo();
        IndigoRenderer renderer = new IndigoRenderer(indigo);
        setRenderingOptions(indigo);

        IndigoObject query = indigo.loadQueryMolecule(highlightedSubstructure);
        IndigoObject mol = highlightSubstructure(indigo, structureData, query);

        byte[] buffer = renderer.renderToBuffer(mol);
        try {
            out.write(buffer);
        } catch (IOException ioEx) {
            throw new MoleculeDatabaseFrameworkException(ioEx);
        }
    }

    /**
     * Creates PNG Image file of the Molecule
     *
     * @param mol
     * @param filePath
     */
    public void renderChemicalStructureToFile(String structureData, String filePath) {

        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(filePath);
        Preconditions.checkArgument(!filePath.isEmpty(),
                "Argument 'filePath' can not be an empty String.");

        Indigo indigo = new Indigo();
        IndigoRenderer renderer = new IndigoRenderer(indigo);
        setRenderingOptions(indigo);

        IndigoObject mol = indigo.loadMolecule(structureData);
        renderer.renderToFile(mol, filePath);
    }

    /**
     * Creates PNG Image file of the Molecule
     *
     * @param mol
     * @param filePath
     */
    public void renderChemicalStructureToFile(String structureData, String filePath,
            String highlightedSubstructure) {

        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(filePath);
        Preconditions.checkArgument(!filePath.isEmpty(),
                "Argument 'filePath' can not be an empty String.");

        Indigo indigo = new Indigo();
        IndigoRenderer renderer = new IndigoRenderer(indigo);
        setRenderingOptions(indigo);

        IndigoObject query = indigo.loadQueryMolecule(highlightedSubstructure);
        IndigoObject mol = highlightSubstructure(indigo, structureData, query);
        renderer.renderToFile(mol, filePath);
    }

    /**
     * <p>Outputs a png image at the given
     * <code>OutputStream out</code>.</p> <p> The images size is according to to
     * the specified with in the constructor. This means the more compositions
     * the
     * <code>compound</code> consists of the smaller each structure will be
     * rendered. However the images size is fixed. </p>
     *
     * @param compound the ChemicalCompound to create an image of
     * @param out the stream to write to
     */
    public void renderChemicalCompound(ChemicalCompound<?> compound,
            OutputStream out) {

        renderChemicalCompound(compound, out, false);
    }

    /**
     * <p>Outputs a png image at the given
     * <code>OutputStream out</code>.</p> <p> If
     * <code>isWidthColumnSize</code> is false, the images size is according to
     * to the specified with in the constructor. This means the more
     * compositions the
     * <code>compound</code> consists of the smaller each structure will be
     * rendered. However the images size is fixed. </p> <p> If
     * <code>isWidthColumnSize</code> is true, each structure gets the specified
     * width. The total with of the image is then
     * <pre>
     * <code>
     *
     * int totalWidth = nrOfCompostions * width;
     *
     * </code>
     * </pre> and hence the images size depends on the input. </p>
     *
     * @param compound the ChemicalCompound to create an image of
     * @param out the stream to write to
     * @param isWidthColumnSize whether width of image is fixed or not
     */
    public void renderChemicalCompound(ChemicalCompound<?> compound,
            OutputStream out, boolean isWidthColumnSize) {

        Preconditions.checkNotNull(compound);
        Preconditions.checkNotNull(out);

        Indigo indigo = new Indigo();
        IndigoRenderer renderer = new IndigoRenderer(indigo);
        setRenderingOptions(indigo);
        IndigoObject arr = indigo.createArray();

        for (ChemicalCompoundComposition composition : compound.getCompositions()) {

            String structureData = composition.getChemicalStructure().getStructureData();
            IndigoObject mol = indigo.loadMolecule(structureData);
            arr.arrayAdd(mol);
        }
        renderCompoundToStream(indigo, renderer, arr, out, isWidthColumnSize);
    }

    /**
     * <p>Outputs a png image at the given
     * <code>OutputStream out</code>.</p><p> If
     * <code>isWidthColumnSize</code> is false, the images size is according to
     * to the specified with in the constructor. This means the more
     * compositions the
     * <code>compound</code> consists of the smaller each structure will be
     * rendered. However the images size is fixed. </p> <p> If
     * <code>isWidthColumnSize</code> is true, each structure gets the specified
     * width. The total with of the image is then
     * <pre>
     * <code>
     *
     * int totalWidth = nrOfCompostions * width;
     *
     * </code>
     * </pre> and hence the images size depends on the input. </p>
     *
     * @param compound the ChemicalCompound to create an image of
     * @param filePath location of the image file
     * @param highlightedSubstructure the part of the contained structures to
     * highlight
     * @param isWidthColumnSize whether width of image is fixed or not
     */
    public void renderChemicalCompound(ChemicalCompound<?> compound,
            OutputStream out, String highlightedSubstructure, boolean isWidthColumnSize) {

        Preconditions.checkNotNull(compound);
        Preconditions.checkNotNull(out);
        Preconditions.checkNotNull(highlightedSubstructure);
        Preconditions.checkArgument(!highlightedSubstructure.isEmpty(),
                "Argument 'highlightedSubstructure' can not be an empty String.");

        Indigo indigo = new Indigo();
        IndigoRenderer renderer = new IndigoRenderer(indigo);
        setRenderingOptions(indigo);
        IndigoObject arr = highlightCompound(indigo, compound,
                highlightedSubstructure);

        renderCompoundToStream(indigo, renderer, arr, out, isWidthColumnSize);
    }

    private void renderCompoundToStream(Indigo indigo, IndigoRenderer renderer,
            IndigoObject arr, OutputStream out, boolean isWidthColumnSize) {

        int nrOfStructures = arr.count();
        if (isWidthColumnSize) {
            int totalWidth = width * nrOfStructures;
            indigo.setOption("render-image-size", totalWidth, height);
        }
        byte[] buffer = renderer.renderGridToBuffer(arr, null, nrOfStructures);
        try {
            out.write(buffer);
        } catch (IOException ioEx) {
            throw new MoleculeDatabaseFrameworkException(ioEx);
        }
    }

    /**
     * <p>Creates a png image at the given location
     * <code>filePath</code>.</p> <p> The images size is according to to the
     * specified with in the constructor. This means the more compositions the
     * <code>compound</code> consists of the smaller each structure will be
     * rendered. However the images size is fixed. </p>
     *
     * @param compound the ChemicalCompound to create an image of
     * @param filePath location of the image file
     */
    public void renderChemicalCompoundToFile(ChemicalCompound<?> compound,
            String filePath) {

        renderChemicalCompoundToFile(compound, filePath, false);
    }

    /**
     * <p>Creates a png image at the given location
     * <code>filePath</code>.</p> <p> If
     * <code>isWidthColumnSize</code> is false, the images size is according to
     * to the specified with in the constructor. This means the more
     * compositions the
     * <code>compound</code> consists of the smaller each structure will be
     * rendered. However the images size is fixed. </p> <p> If
     * <code>isWidthColumnSize</code> is true, each structure gets the specified
     * width. The total with of the image is then
     * <pre>
     * <code>
     *
     * int totalWidth = nrOfCompostions * width;
     *
     * </code>
     * </pre> and hence the images size depends on the input. </p>
     *
     * @param compound the ChemicalCompound to create an image of
     * @param filePath location of the image file
     * @param isWidthColumnSize whether width of image is fixed or not
     */
    public void renderChemicalCompoundToFile(ChemicalCompound<?> compound,
            String filePath, boolean isWidthColumnSize) {

        Preconditions.checkNotNull(compound);
        Preconditions.checkNotNull(filePath);
        Preconditions.checkArgument(!filePath.isEmpty(),
                "Argument 'filePath' can not be an empty String.");

        Indigo indigo = new Indigo();
        IndigoRenderer renderer = new IndigoRenderer(indigo);
        setRenderingOptions(indigo);
        IndigoObject arr = indigo.createArray();

        for (ChemicalCompoundComposition composition : compound.getCompositions()) {

            String structureData = composition.getChemicalStructure().getStructureData();
            IndigoObject mol = indigo.loadMolecule(structureData);
            arr.arrayAdd(mol);
        }

        renderCompoundToFile(indigo, renderer, arr, filePath, isWidthColumnSize);
    }

    /**
     * <p> Creates a png image at the given location
     * <code>filePath</code>. </p><p> If
     * <code>isWidthColumnSize</code> is false, the images size is according to
     * to the specified with in the constructor. This means the more
     * compositions the
     * <code>compound</code> consists of the smaller each structure will be
     * rendered. However the images size is fixed. </p> <p> If
     * <code>isWidthColumnSize</code> is true, each structure gets the specified
     * width. The total with of the image is then
     * <pre>
     * <code>
     *
     * int totalWidth = nrOfCompostions * width;
     *
     * </code>
     * </pre> and hence the images size depends on the input. </p>
     *
     * @param compound the ChemicalCompound to create an image of
     * @param filePath location of the image file
     * @param highlightedSubstructure the part of the contained structures to
     * highlight
     * @param isWidthColumnSize whether width of image is fixed or not
     */
    public void renderChemicalCompoundToFile(ChemicalCompound<?> compound,
            String filePath, String highlightedSubstructure, boolean isWidthColumnSize) {

        Preconditions.checkNotNull(compound);
        Preconditions.checkNotNull(filePath);
        Preconditions.checkNotNull(highlightedSubstructure);
        Preconditions.checkArgument(!filePath.isEmpty(),
                "Argument 'filePath' can not be an empty String.");
        Preconditions.checkArgument(!highlightedSubstructure.isEmpty(),
                "Argument 'highlightedSubstructure' can not be an empty String.");

        Indigo indigo = new Indigo();
        IndigoRenderer renderer = new IndigoRenderer(indigo);
        setRenderingOptions(indigo);
        IndigoObject arr = highlightCompound(indigo, compound,
                highlightedSubstructure);

        renderCompoundToFile(indigo, renderer, arr, filePath, isWidthColumnSize);
    }

    private void renderCompoundToFile(Indigo indigo, IndigoRenderer renderer,
            IndigoObject arr, String filePath, boolean isWidthColumnSize) {

        int nrOfStructures = arr.count();
        if (isWidthColumnSize) {
            int totalWidth = width * nrOfStructures;
            indigo.setOption("render-image-size", totalWidth, height);
        }
        renderer.renderGridToFile(arr, null, nrOfStructures, filePath);
    }

    private IndigoObject highlightCompound(Indigo indigo,
            ChemicalCompound<?> compound, String highlightedSubstructure) {

        IndigoObject arr = indigo.createArray();
        IndigoObject query = indigo.loadQueryMolecule(highlightedSubstructure);
        for (ChemicalCompoundComposition composition : compound.getCompositions()) {
            String structureData = composition.getChemicalStructure().getStructureData();
            IndigoObject mol = highlightSubstructure(indigo, structureData, query);
            arr.arrayAdd(mol);
        }
        return arr;
    }

    private IndigoObject highlightSubstructure(Indigo indigo,
            String structureData, IndigoObject query) {

        IndigoObject mol = indigo.loadMolecule(structureData);
        IndigoObject matcher = indigo.substructureMatcher(mol);

        IndigoObject match = matcher.match(query);
        if (match != null) {
            for (IndigoObject item : query.iterateAtoms()) {
                IndigoObject atom = match.mapAtom(item);
                atom.highlight();

                for (IndigoObject nei : atom.iterateNeighbors()) {
                    if (!nei.isPseudoatom() && !nei.isRSite() && nei.atomicNumber() == 1) {
                        nei.highlight();
                        nei.bond().highlight();
                    }
                }
            }

            for (IndigoObject bond : query.iterateBonds()) {
                match.mapBond(bond).highlight();
            }
        }
        return mol;
    }

    /**
     * Creates a renderer with default options set.
     *
     * @param indigo
     * @return
     */
    private void setRenderingOptions(Indigo indigo) {
        indigo.setOption("render-output-format", "png");
        indigo.setOption("render-margins", 5, 5);
        indigo.setOption("render-image-size", width, height);
        indigo.setOption("render-coloring", true);
        indigo.setOption("render-stereo-style", "none");
    }
}
