/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import com.google.common.primitives.Primitives;
import java.lang.reflect.Field;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.util.ReflectionUtils;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class PropertyGetter {

    private static final XLogger logger = XLoggerFactory.getXLogger("PropertyGetter");
    private static final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";

    /**
     * <p>
     * Generates a mapping of property name (field name) to the properties value
     * as String. Only primitives and Date/Calendar fields are considered.
     * </p>
     * <p>
     * The Map includes all properties that are contained in <code>propertyNames</code>.
     * For date fields default format of "yyyy-MM-dd HH:mm:ss" is used.
     * </p>
     * <p>
     * <code>Object entity</code> must be an instance of <code>Class entityClass</code>.
     * </p>
     *
     * @param entity
     * @param entityClass
     * @param propertyNames
     * @return
     * @throws NoSuchFieldException if a given field with given name does not exist.
     */
    public static Map<String, String> getProperties(Object entity, Class entityClass,
            Set<String> propertyNames)  throws NoSuchFieldException {
        return getProperties(entity, entityClass, propertyNames, DATE_FORMAT);
    }

    /**
     * <p>
     * Generates a mapping of property name (field name) to the properties value
     * as String. Only primitives and Date/Calendar fields are considered.
     * </p>
     * <p>
     * The Map includes all properties that are contained in <code>propertyNames</code>.
     * For date fields the date format defined by <code>dateFormat</code> is used.
     * See <a href="http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html">SimpleDateFormat</a>
     * for more information.
     * </p>
     * <p>
     * <code>Object entity</code> must be an instance of <code>Class entityClass</code>.
     * </p>
     *
     * @param entity
     * @param entityClass
     * @param propertyNames
     * @param dateFormat
     * @return
     * @throws NoSuchFieldException if a given field with given name does not exist.
     * @throws IllegalArgumentException if <code>dateFormat</code> is invalid
     */
    public static Map<String, String> getProperties(Object entity, Class entityClass,
            Set<String> propertyNames, String dateFormat)
            throws NoSuchFieldException {

        logger.entry(entity, entityClass, propertyNames);
        Map<String, String> values = new HashMap<>();
        SimpleDateFormat formatter =
                    new SimpleDateFormat(dateFormat);

        if (propertyNames != null) {
            for (String fieldName : propertyNames) {

                //Get field recursively up the inheritance "tree" up to object
                Field field = ReflectionUtils.findField(entityClass, fieldName);
                if (field == null) {
                    throw new NoSuchFieldException(String.format("No Field %s found in Class %s.", fieldName, entityClass.getSimpleName()));
                }

                String value = getValueAsString(entity, field, formatter);
                values.put(fieldName, value);
            }
        }
        logger.exit(values);
        return values;
    }

    private static String getValueAsString(Object entity, Field field,
            SimpleDateFormat formatter) {

        Class fieldType = field.getType();
        logger.debug("Trying to get Field {} of type {}.", field.getName(), fieldType.getName());
        ReflectionUtils.makeAccessible(field);
        String value = "";

        if (fieldType.isPrimitive()|| Primitives.isWrapperType(fieldType)) {
            value = ReflectionUtils.getField(field, entity).toString();
        } else if (fieldType.equals(String.class)) {
            value = (String) ReflectionUtils.getField(field, entity);
            if(value == null){
                value = "";
            }

        } else if (fieldType.equals(Date.class)
                || fieldType.equals(Calendar.class)) {

            if (fieldType.equals(Date.class)) {
                Date date = (Date) ReflectionUtils.getField(field, entity);
                if (date == null){
                    value = "";
                } else {
                    value = formatter.format(date);
                }
            } else {
                Calendar cal = (Calendar) ReflectionUtils.getField(field, entity);
                Date date = cal.getTime();
                if (date == null){
                    value = "";
                } else {
                    value = formatter.format(date);
                }
            }
        }
        return value;
    }
}
