/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import java.text.ParseException;
import java.util.Map;
import java.util.concurrent.Callable;
import lombok.AllArgsConstructor;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalCompoundFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.FormatConversionException;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;

/**
 * <p>
 * Imports a single <code>SdfRecord</code> into the database.
 * </p>
 * <p>
 * See {@link org.bitbucket.kienerj.moleculedatabaseframework.service.ChemicalCompoundService
 * #importSDF(Reader,Map<String, String>,Map<String,Map<String,Double>>,
            String, boolean) chemicalCompoundService.importSDF}
 * for further information.
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@AllArgsConstructor
public class SdfCompoundCreator<T extends ChemicalCompound>
        implements Callable<T> {

    private static final XLogger logger = XLoggerFactory.getXLogger("SdfCompoundImporter");

    private final SdfRecord sdfRecord;
    private final ChemicalCompoundFactory<T> chemicalCompoundFactory;
    private final Map<String, String> mappedProperties;
    private final Map<String,Double> percentages;
    private final EntityImportResult importResult;
    private final boolean abortOnError;

    @Override
    public T call() throws ParseException,
            IllegalAccessException, NoSuchFieldException, InstantiationException {

        logger.entry();
        try {
            T chemicalCompound = chemicalCompoundFactory
                    .createChemicalCompound(sdfRecord, percentages, mappedProperties);
            logger.exit(chemicalCompound);
            return chemicalCompound;

        } catch (FormatConversionException ex) {
            if (abortOnError) {
                throw ex;
            } else {
                CompoundImportError importError = new CompoundImportError(
                        ex.getMessage(), ex.getStructureData());
                importResult.addImportError(importError);
                logger.warn("Found invalid molecule with molfilename {} while importing SD-File.", sdfRecord.getMolfileName());
                return null;
            }
        }
    }
}
