/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import java.text.ParseException;
import java.util.Map;
import java.util.concurrent.Callable;
import lombok.AllArgsConstructor;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalCompoundFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.FormatConversionException;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@AllArgsConstructor
public class SdfContainableCreator<T extends Containable>
        implements Callable<T>{

    private static final XLogger logger = XLoggerFactory.getXLogger("SdfContainableImporter");

    private final SdfRecord sdfRecord;
    private final ChemicalCompoundFactory chemicalCompoundFactory;
    private final ContainableFactory<T> containableFactory;
    private final Map<String, String> mappedCompoundProperties;
    private final Map<String, String> mappedContainableProperties;
    private final Map<String, Double> percentages;
    private ChemicalCompound compound;
    private final EntityImportResult importResult;
    private final boolean abortOnError;

    @Override
    public T call() throws ParseException,
            IllegalAccessException, NoSuchFieldException, InstantiationException,
            NoSuchMethodException {

        logger.entry();

        try {
            //if compound is null it does not exist yet, hence create
            if (compound == null) {
                compound = chemicalCompoundFactory
                        .createChemicalCompound(sdfRecord, percentages, mappedCompoundProperties);
            }

            T  containable = containableFactory
                        .createContainable(compound, mappedContainableProperties, sdfRecord.getProperties());

            logger.exit(containable);
            return containable;

        } catch (FormatConversionException ex) {
            if (abortOnError) {
                throw ex;
            } else {
                CompoundImportError importError = new CompoundImportError(
                        ex.getMessage(), ex.getStructureData());
                importResult.addImportError(importError);
                logger.warn("Found invalid molecule with molfilename {} while importing SD-File.", sdfRecord.getMolfileName());
                return null;
            }
        }
    }
}
