/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.repository;

import java.util.List;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.springframework.data.jpa.repository.Query;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

/**
 * Spring-Data Repository for <code>ChemicalStructure</code> entity.
 *
 * IMPORTANT: Never call save() method from this repository directly. Always use
 * the Service <code>ChemicalStructureService</code>.
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Transactional(propagation = Propagation.MANDATORY)
@Repository("chemicalStructureRepository")
public interface ChemicalStructureRepository<T extends ChemicalStructure>
        extends BaseEntityRepository<T> {

    /**
     * Gets a structure by its unique key
     *
     * @param structureKey
     * @return
     */
    T findByStructureKey(String structureKey);

    /**
     * Gets all compounds a given structure appears in.
     *
     * @param structure
     * @return
     */
    @Query("select c from ChemicalCompound c join c.compositions comp where comp.pk.chemicalStructure = ?1")
    List<ChemicalCompound> getChemicalCompoundsByStructure(ChemicalStructure structure);

    /**
     * Gets all compounds a given structure appears in.
     *
     * @param structureId
     * @return
     */
    @Query("select c from ChemicalCompound c join c.compositions comp where comp.pk.chemicalStructure.id = ?1")
    List<ChemicalCompound> getChemicalCompoundsByStructureId(Long structureId);

    /**
     * Gets all <code>ChemicalCompoundCompositions</code> a given <code>ChemicalStructure</code>
     * appears in.
     *
     * @param structure
     * @return
     */
    @Query("select comp from ChemicalCompoundComposition comp where comp.pk.chemicalStructure = ?1")
    List<ChemicalCompoundComposition> getCompositionsByStructure(ChemicalStructure structure);

    /**
     * Gets all <code>ChemicalCompoundCompositions</code> a given <code>ChemicalStructure</code>
     * appears in.
     *
     * @param structureId
     * @return
     */
    @Query("select comp from ChemicalCompoundComposition comp where comp.pk.chemicalStructure.id = ?1")
    List<ChemicalCompoundComposition> getCompositionsByStructureId(Long structureId);

    /**
     *
     * @param structureData SMILES or molfile
     * @return
     */
    @Query(value="SELECT Bingo.InChI(?1, '')", nativeQuery=true)
    String getStandardInChi(String structureData);

    /**
     *
     * @param structureData SMILES or molfile
     * @return
     */
    @Query(value="SELECT Bingo.InChIKey(Bingo.InChI(?1, ''))", nativeQuery=true)
    String getStandardInChiKey(String structureData);


    /**
     * Gets an InChi. If options is null or empty string, the result will
     * be the same as <code>getStandardInChi(structureData)</code>
     *
     * See <a href="http://www.inchi-trust.org/fileadmin/user_upload/html/inchifaq/inchi-faq.html#15.17">Information on InChi options</a>.
     *
     * @param structureData SMILES or molfile
     * @param options options for generating InChi
     * @return
     */
    @Query(value="SELECT Bingo.InChI(?1, ?2)", nativeQuery=true)
    String getInChi(String structureData, String options);

    /**
     * Gets an InChi key. If options is null or empty string, the result will
     * be the same as <code>getStandardInChiKey(structureData)</code>
     *
     * See <a href="http://www.inchi-trust.org/fileadmin/user_upload/html/inchifaq/inchi-faq.html#15.17">Information on InChi options</a>.
     *
     * @param structureData SMILES or molfile
     * @param options options for generating InChi
     * @return
     */
    @Query(value="SELECT Bingo.InChIKey(Bingo.InChI(?1, ?2))", nativeQuery=true)
    String getInChiKey(String structureData, String options);

}
