/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.repository;

import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.path.PathBuilder;
import java.util.List;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.SimilarityType;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.repository.NoRepositoryBean;

/**
 * <p> Interface for enabling chemical structure searching for implementing
 * repository. </p> <p> Method Usage Explanation: </p> <p> Page size and sorting
 * is done based in pageable. Example: </p>
 * <pre>
 * <code>
 * PageRequest pageable = new PageRequest( offset, limit,
 *         new Sort(
 *             new Order(Direction.ASC, column1),
 *             new Order(Direction.ASC, column2)
 *         )
 * );
 * </code>
 * </pre> <p>
 * <code>
 * PathBuilder<T> pathBuilder = new PathBuilder<>(T.class, "classSimpleName");
 * </code> Here T needs to be the actual ClassName and not a generic type. </p>
 * <p>
 * <code>predicate</code> is a querydsl Predicate. This can be used to apply
 * additional search restrictions, eg. the results must match by chemical
 * structure and the predicate. </p> <p> A Predicate is created by using the
 * querydsl generated query classes. </p> <p> Example: </p>
 * <pre>
 * <code>
 *
 * QChemicalCompound compound = QChemicalCompound.chemicalCompound;
 * QChemicalCompoundComposition composition =
 *           QChemicalCompoundComposition.chemicalCompoundComposition;
 *
 * Predicate predicate =
 *     compound.compoundName.like("Methyl%").and(composition.percentage.gt(50));
 *
 * </code>
 * </pre> <p> And at last searchOptions is a String defining how the Bingo
 * cartridge should perform the search. See
 * http://ggasoftware.com/opensource/bingo/user-manual-postgres#substructure-search
 * for further information. Using an empty string is the default and in most
 * cases right choice unless you need special types of searches. </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@NoRepositoryBean
public interface ChemicalStructureSearchRepository<T> {

    /**
     * <p> Returns a Page of entities matching the chemical structure and other
     * query predicates (filters). </p>
     *
     * <p> Page size and sorting is done based in pageable. Example: </p>
     * <pre>
     * <code>
     * PageRequest pageable = new PageRequest( offset, limit,
     *         new Sort(
     *             new Order(Direction.ASC, column1),
     *             new Order(Direction.ASC, column2)
     *         )
     * );
     * </code>
     * </pre> <p>
     * <code>
     * PathBuilder<T> pathBuilder = new PathBuilder<>(T.class, "classSimpleName");
     * </code> Here T needs to be the actual ClassName and not a generic type.
     * </p> <p>
     * <code>predicate</code> is a querydsl Predicate. This can be used to apply
     * additional search restrictions, eg. the results must match by chemical
     * structure and the predicate. </p> <p> A Predicate is created by using the
     * querydsl generated query classes. </p> <p> Example: </p>
     * <pre>
     * <code>
     *
     * QChemicalCompound compound = QChemicalCompound.chemicalCompound;
     * QChemicalCompoundComposition composition =
     *           QChemicalCompoundComposition.chemicalCompoundComposition;
     *
     * Predicate predicate =
     *     compound.compoundName.like("Methyl%").and(composition.percentage.gt(50));
     *
     * </code>
     * </pre> <p> And at last searchOptions is a String defining how the Bingo
     * cartridge should perform the search. See
     * http://ggasoftware.com/opensource/bingo/user-manual-postgres#substructure-search
     * for further information. Using an empty string is the default and in most
     * cases right choice unless you need special types of searches. </p>
     *
     * @param structureData
     * @param searchType
     * @param pathBuilder
     * @param pageable
     * @param predicate
     * @param searchOptions
     * @return
     */
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions,
            PathBuilder<T> pathBuilder);

    /**
     * <p>See method <code>Page<T> findByChemicalStructure()</code></p>.
     *
     * <p>
     * This method returns a List of ids of all matching <code>ChemicalCompounds</code>
     * instead of entity instances. The list can be used for exporting matches
     * to sdf, as example.
     * </p>
     *
     * @param structureData
     * @param searchType
     * @param predicate
     * @param searchOptions
     * @param compoundPathBuilder
     * @param orders
     * @return
     */
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate, String searchOptions,
            PathBuilder<T> compoundPathBuilder, OrderSpecifier<?>... orders);

    /**
     * <p> Returns a Page of entities similar to the searched chemical structure
     * and matching the other query predicates (filters). </p>
     *
     * * <p> Page size and sorting is done based in pageable. Example: </p>
     * <pre>
     * <code>
     * PageRequest pageable = new PageRequest( offset, limit,
     *         new Sort(
     *             new Order(Direction.ASC, column1),
     *             new Order(Direction.ASC, column2)
     *         )
     * );
     * </code>
     * </pre> <p>
     * <code>
     * PathBuilder<T> pathBuilder = new PathBuilder<>(T.class, "classSimpleName");
     * </code> Here T needs to be the actual ClassName and not a generic type.
     * </p> <p>
     * <code>predicate</code> is a querydsl Predicate. This can be used to apply
     * additional search restrictions, eg. the results must match by chemical
     * structure and the predicate. </p> <p> A Predicate is created by using the
     * querydsl generated query classes. </p> <p> Example: </p>
     * <pre>
     * <code>
     *
     * QChemicalCompound compound = QChemicalCompound.chemicalCompound;
     * QChemicalCompoundComposition composition =
     *           QChemicalCompoundComposition.chemicalCompoundComposition;
     *
     * Predicate predicate =
     *     compound.compoundName.like("Methyl%").and(composition.percentage.gt(50));
     *
     * </code>
     * </pre>
     * <p>
     * lowerBound and upperBound define the similarity threshold values. They can range
     * from 0 to 1. similarityType defines the algorithm to use to measure the
     * similarity.
     * </p>
     *
     * @param structureData
     * @param similarityType
     * @param lowerBound
     * @param upperBound
     * @param pathBuilder
     * @param pageable
     * @param predicate
     * @return
     */
    Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate,
            PathBuilder<T> pathBuilder);

    /**
     * <p> Returns a List of ids of entities similar to the searched chemical
     * structure and matching the other query predicates (filters). </p>
     *
     * *<p> Sorting is done based on QueryDSL OrderSpecifiers.</p>
     * <p>
     * <code>
     * PathBuilder<T> pathBuilder = new PathBuilder<>(T.class, "classSimpleName");
     * </code> Here T needs to be the actual ClassName and not a generic type.
     * </p> <p>
     * <code>predicate</code> is a QueryDSL Predicate. This can be used to apply
     * additional search restrictions, eg. the results must match by chemical
     * structure and the predicate. </p> <p> A Predicate is created by using the
     * QueryDSL generated query classes. </p> <p> Example: </p>
     * <pre>
     * <code>
     *
     * QChemicalCompound compound = QChemicalCompound.chemicalCompound;
     * QChemicalCompoundComposition composition =
     *           QChemicalCompoundComposition.chemicalCompoundComposition;
     *
     * Predicate predicate =
     *     compound.compoundName.like("Methyl%").and(composition.percentage.gt(50));
     *
     * </code>
     * </pre>
     * <p>
     * lowerBound and upperBound define the similarity threshold values. They can range
     * from 0 to 1. similarityType defines the algorithm to use to measure the
     * similarity.
     * </p>
     *
     * @param structureData
     * @param similarityType
     * @param lowerBound
     * @param upperBound
     * @param predicate
     * @param compoundPathBuilder
     * @param orders
     * @return
     */
    List<Long> findBySimilarStructure(String structureData,
            SimilarityType similarityType, Double lowerBound, Double upperBound,
            Predicate predicate, PathBuilder<T> compoundPathBuilder,
            OrderSpecifier<?>... orders);
}
