/*
 * Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.security;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.hierarchicalroles.RoleHierarchy;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.userdetails.UserDetails;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class ChemicalCompoundPermission extends AbstractPermission {

    private static final String INVALID_PERMISSION_MESSAGE = "%s is not a valid permission type. "
            + "Permission Type must be 'save' or 'delete'.";

    public ChemicalCompoundPermission() {
        super("ChemicalCompound");
    }

    @Override
    public boolean isAllowed(Authentication authentication, Object targetDomainObject,
            String permissionType, String permissionTarget) {

        if (!isAuthenticated(authentication)) {
            return false;
        }

        boolean hasPermission = false;
        switch (permissionType) {
            case "save":
                hasPermission = canSave(authentication, (ChemicalCompound) targetDomainObject, permissionTarget);
                break;
            case "delete":
                hasPermission = canDelete(authentication, (ChemicalCompound) targetDomainObject, permissionTarget);
                break;
            default:
                throw new InvalidPermissionException(String.format(INVALID_PERMISSION_MESSAGE, permissionType));
        }
        return hasPermission;
    }

    /**
     * save => create or update
     *
     */
    private boolean canSave(final Authentication authentication,
            final ChemicalCompound targetDomainObject,
            final String permissionTarget) {
        final boolean isNew = targetDomainObject.getId() == null;

        final boolean canSave = (hasRole(authentication, "create_" + permissionTarget) && isNew)
                || (hasRole(authentication, "update_created_" + permissionTarget)
                && isCreator(authentication, targetDomainObject))
                || hasRole(authentication, "update_" + permissionTarget);

        if (!canSave) {
            return false;
        }

        /*
         * Relationship of entities have CascadeType.PERSIST set but not .MERGE hence
         * if targetDomainObject is a new entity we need to check if current
         * user has privilege to create any of the associated entites in targetDomainObject.
         * associated entites -> Containables and Containers in Containables
         *
         * Since CascadeType.MERGE is not set, changes to associated entites
         * are ignored when updating and hence no need to check if user can update them
         * (Checking if entity has changed is IMHO notn-trivial here in this context
         * hence this design)
         */
        if (isNew && !targetDomainObject.getContainables().isEmpty()
                && !canCreateContainablesAndContainers(authentication, targetDomainObject)) {
            return false;
        }

        return canSave;
    }

    private boolean canDelete(final Authentication authentication,
            final ChemicalCompound targetDomainObject,
            final String permissionTarget) {
        return (hasRole(authentication, "delete_created_" + permissionTarget)
                && isCreator(authentication, targetDomainObject))
                || hasRole(authentication, "delete_" + permissionTarget);
    }

    /**
     * <p>
     * Checks if current user has privilege to create new Containables and
     * Containers associated with the ChemicalCompound
     * <code>targetDomainObject</code>.
     * </p>
     * <p>
     * Always returns true if
     *
     * <ocde>targetDomainObject.getContainables.isEmpty() == true</code>
     * </p>
     * @param authentication
     * @param targetDomainObject
     * @return
     */
    private boolean canCreateContainablesAndContainers(final Authentication authentication,
            final ChemicalCompound targetDomainObject) {

        boolean canSaveContainables = true;
        final Iterator<Containable> it = ((Set<Containable>) targetDomainObject.getContainables()).iterator();

        //iterator & while loops: so the loops can be left as soon as canSaveContainables == false
        while (canSaveContainables && it.hasNext()) {
            final Containable containable = it.next();
            final String containableClassName = containable.getClass().getSimpleName();
            final boolean isNewContainable = containable.getId() == null;
            canSaveContainables = canSaveContainables && (hasRole(authentication, "create_" + containableClassName) && isNewContainable);

            Iterator<ChemicalCompoundContainer> containerIterator = containable.getContainers().iterator();
            while (canSaveContainables && containerIterator.hasNext()) {
                final ChemicalCompoundContainer container = containerIterator.next();
                final String containerClassName = container.getClass().getSimpleName();
                final boolean isNewContainer = container.getId() == null;
                canSaveContainables = canSaveContainables && (hasRole(authentication, "create_" + containerClassName) && isNewContainer);
            }
        }
        return canSaveContainables;
    }
}
