/*
 * Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.security;

import com.google.common.base.Preconditions;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import org.springframework.security.access.PermissionEvaluator;
import org.springframework.security.core.Authentication;
import org.springframework.util.StringUtils;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class DefaultPermissionEvaluator implements PermissionEvaluator {

    private Map<String, Permission> permissionNameToPermissionMap = new HashMap<>();


    private static final String INVALID_PERMISSION_MESSAGE = "%s is not a valid permission. "
                + "It must contain 1 underscore like in save_TestCompound.";

    protected DefaultPermissionEvaluator(){}

    public DefaultPermissionEvaluator(Map<String, Permission> permissionNameToPermissionMap) {
        Preconditions.checkNotNull(permissionNameToPermissionMap);
        this.permissionNameToPermissionMap = permissionNameToPermissionMap;
    }

    @Override
    public boolean hasPermission(Authentication authentication, Object targetDomainObject, Object permission) {

        boolean hasPermission = false;
        if (canHandleByDomainObject(authentication, targetDomainObject, permission)) {
            hasPermission = checkPermissionByDomainObject(authentication, targetDomainObject, (String) permission);
        }
        return hasPermission;
    }

    @Override
    public boolean hasPermission(Authentication authentication, Serializable targetId, String targetType, Object permission) {
        // due to a bug in Spring Data, you can't autowire repositories (or services containing them)
        // in Spring Security (Spring context won't load). Therefore we can't access the actual
        // object using the id to check who the creator is (as example).
        // -> access per id never allowed.
        return false;
    }


    private boolean canHandleByDomainObject(Authentication authentication,
            Object targetDomainObject, Object permission) {
        return targetDomainObject != null && authentication != null && permission instanceof String;
    }

    private boolean checkPermissionByDomainObject(Authentication authentication,
            Object targetDomainObject, String permissionValue) {
        String permissionType = getPermissionType(permissionValue);
        String permissionTarget = getPermissionTarget(permissionValue);
        verifyPermissionIsDefined(permissionTarget);
        Permission permission = permissionNameToPermissionMap.get(permissionTarget);
        return permission.isAllowed(authentication, targetDomainObject, permissionType, permissionTarget);
    }

    /**
     * example: create_TestCompound -> returns create
     *
     * @param permission
     * @return
     */
    private String getPermissionType(String permission){
        String[] result =  permission.split("_");

        int underscoreCount = StringUtils.countOccurrencesOf(permission, "_");

        if (underscoreCount == 1){
            return result[0];
        } else {
            throw new InvalidPermissionException(String.format(INVALID_PERMISSION_MESSAGE, permission));
        }
    }

    /**
     * example: create_TestCompound -> returns TestCompound
     *
     * @param permission
     * @return
     */
    private String getPermissionTarget(String permission){
        String[] result =  permission.split("_");

        int underscoreCount = StringUtils.countOccurrencesOf(permission, "_");

        if (underscoreCount == 1){
            return result[1];
        } else {
            throw new InvalidPermissionException(String.format(INVALID_PERMISSION_MESSAGE, permission));
        }
    }

    private void verifyPermissionIsDefined(String permissionKey) {
        if (!permissionNameToPermissionMap.containsKey(permissionKey)) {
            throw new PermissionNotDefinedException("No permission with key " + permissionKey + " is defined in " + this.getClass().toString());
        }
    }
}
