/*
 * Copyright (C) 2012 Joos Kiener
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.path.PathBuilder;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.io.EntityImportResult;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfRecord;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.SimilarityType;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p>
 * Service interface for data access of any type of <code>ChemicalCompound</code>.
 * This includes methods for saving, retrieving and searching for <code>ChemicalCompounds</code>.
 * </p>
 * <p>
 * All service methods are transactional and annotated with Spring Security. Security
 * must be explictly enabled in configuration.
 * </p>
 * <p>
 * To create a service for your <code>ChemicalCompound</code> implementations,
 * your need to extend this interface and create an implementation extending
 * <code>ChemicalCompoundServiceImpl</code> and implementing your service interface.
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Transactional(readOnly=true)
public interface ChemicalCompoundService<T extends ChemicalCompound>
        extends Service<T> {

    /**
     * Gets a compound by its internal id with lazy collections loaded.
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param id unique identifier for entity to retrieve. Must not be null
     * @return the ChemicalCompound that has the given id
     *
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    @Override
    T getById(final Long id);

    /**
     * Gets a compound by its internal id.
     *
     * <p>
     * loadCompositions and loadContainables determine if the according Lazy collections
     * in ChemicalCompound should be loaded or not.
     * </p>
     * <p>
     * In case of false <code>chemicalCompound.getCompostions()</code> or
     * <code>chemicalCompound.getContainables()</code> will cause a
     * LazyInitializationException outside of a EntityManager session.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param id unique identifier for entity to rertrieve. Must not be null
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @return the ChemicalCompound that has the given id
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T getById(Long id, boolean loadCompositions, boolean loadContainables);

    /**
     * Gets all compounds with an id that is contained in
     * <code>ids</code>.
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param ids unique identifiers for entities to rertrieve. Must not be null
     * @return a <code>List</code> of all the entites found.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> getById(final List<Long> ids);


    /**
     * <p>
     * Gets all compounds with an id that is contained in
     * <code>ids</code>.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param ids unique identifiers for entities to rertrieve. Must not be null
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @return a <code>List</code> of all the entites found.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> getById(List<Long> ids, boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Get compound by cas number
     * </p/>
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param cas the CAS number of the Compound to retrieve. Must not be null.
     * @return the <code>ChemicalCompound</code> with CAS number <code>cas</code> or null
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T getByCas(String cas);

    /**
     * <p>
     * Get compound by cas number.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param cas the CAS number of the Compound to retrieve. Must not be null.
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @return the <code>ChemicalCompound</code> with CAS number <code>cas</code> or null
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T getByCas(String cas, boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Returns an unsorted List of all compounds id matching the predicate.
     * </p>     *
     * <p>
     * <code>predicate</code> must not be null or an NullPointertException is thrown.
     * </p>     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compounds to return. must not be null.
     * @return a <code>List</code> of all compounds ids matching the given <code>predicate</code>
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> getAllIds(Predicate predicate);

    /**
     * <p>
     * Get all <code>ChemicalCompoundCompositions</code> of which <code>compound</code> consists.
     * </p>
     * <p>
     * This is required if lazy collection <code>compositions</code> was not
     * initialized when loading the <code>compound</code>.
     * </p>
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param compound to get compositions from. Must not be null.
     * @return all <code>ChemicalCompoundCompositions compound</code> is made of.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<ChemicalCompoundComposition> getCompositions(T compound);

    /**
     * <p>
     * Get all <code>ChemicalCompoundCompositions</code> of a <code>ChemicalCompound</code> with
     * <code>compoundId</code>.
     * </p>
     * <p>
     * This is required if lazy collection <code>compositions</code> was not
     * initialized when loading the <code>ChemicalCompound</code>.
     * </p>
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param compoundId the compounds id for which to get compositions. Must not be null.
     * @return all <code>ChemicalCompoundCompositions</code> of <code>compound</code> with id <code>compoundId</code>.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<ChemicalCompoundComposition> getCompositions(Long compoundId);

    /**
     * Returns 1 compound matching the given predicate.
     *
     * <p>
     * <code>predicate</code> must not be null or an NullPointertException is thrown.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compound to return. must not be null
     * @return the first matching compound
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T findOne(Predicate predicate);

    /**
     * <p>
     * Returns 1 compound matching the given predicate.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * <p>
     * <code>predicate</code> must not be null or an NullPointertException is thrown.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compound to return. must not be null.
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @return the first matching compound
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T findOne(Predicate predicate, boolean loadCompositions, boolean loadContainables);


    /**
     * <p>
     * Returns an unsorted List of all compounds matching the predicate.
     * </p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * </p>
     *
     * <p>
     * <code>predicate</code> must not be null or an NullPointertException is thrown.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compounds to return. must not be null.
     * @return a <code>List</code> of all compounds matching the given <code>predicate</code>
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate);

    /**
     * <p>
     * Returns an unsorted List of all compounds matching the predicate.
     * </p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * </p>
     *
     * <p>
     * <code>predicate</code> must not be null or an NullPointertException is thrown.
     * </p>
     *
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compounds to return. must not be null.
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @return a <code>List</code> of all compounds matching the given <code>predicate</code>
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate, boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Returns a sorted List of all compounds matching the predicate.
     *</p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     *</p>
     *
     * <p>
     * <code>predicate</code> must not be null or an IllegalArgumentException is thrown.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compounds to return. must not be null.
     * @param orders how the result should be sorted.
     * @return a <code>List</code> of all compounds matching the given <code>predicate</code> and sorted in the requested manner.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate, OrderSpecifier<?>... orders);

    /**
     * <p>
     * Returns a sorted List of all compounds matching the predicate.
     *</p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * </p>
     * <p>
     * <code>predicate</code> must not be null or an IllegalArgumentException is thrown.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compounds to return. must not be null.
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @param orders how the result should be sorted.
     * @return a <code>List</code> of all compounds matching the given <code>predicate</code> and sorted in the requested manner.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate, boolean loadCompositions, boolean loadContainables, OrderSpecifier<?>... orders);

    /**
     * <p>
     * Returns a Page of compounds matching the given predicate.
     * </p>
     * <p>
     * If <code>pageable</code> contained sorting information the returned <code>Page</code>
     * is sorted in that way. The page contains the records as specified in <code>pageable</code>,
     * meaning the first record to return and how many (offset and limit).
     * </p>
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compounds to return.
     * @param pageable contains offset, limit and sorting information. Must not be null.
     * @return a <code>Page</code> of compounds matching the given <code>predicate</code> and sorted in the requested manner.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findAll(Predicate predicate, Pageable pageable);

    /**
     * <p>
     * Returns a Page of compounds matching the given predicate.
     * </p>
     * <p>
     * If <code>pageable</code> contained sorting information the returned <code>Page</code>
     * is sorted in that way. The page contains the records as specified in <code>pageable</code>,
     * meaning the first record to return and how many (offset and limit).
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     * <p>
     * <code>predicate</code> must not be null or an IllegalArgumentException is thrown.
     * </p>
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which compounds to return.
     * @param pageable contains offset, limit and sorting information. Must not be null.
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @return a <code>Page</code> of compounds matching the given <code>predicate</code> and sorted in the requested manner.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findAll(Predicate predicate, Pageable pageable, boolean loadCompositions, boolean loadContainables);


    /**
     * <p>
     * Returns a Page of compounds that contain at least 1 chemical structure
     * matching the chemical structure search arguments.
     * </p>
     * <p>
     * <code>structureData</code> must be in SMILES or molFile format.
     * </p>
     * <p>Example of Pageable:</p>
     * <code>
     * PageRequest pageable = new PageRequest(
     *           offset, limit, new Sort(
     *           new Order(Direction.ASC, "compoundName")));
     * </code>
     * <p>
     * The name given to Order ("compoundName") must equal a field name
     * of ChemicalCompound class.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param structureData the chemical structure to search for
     * @param searchType FullStructure, SubStructure, SMARTS or Formula search
     * @param pageable contains offset, limit and sorting information. Must nto be null.
     * @return a <code>Page</code> of compounds matching the given <code>structureData</code>.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable);

    /**
     * <p>
     * Returns a Page of compounds that contain at least 1 chemical structure
     * matching the chemical structure search arguments and the predicate.
     * </p>
     * <p>
     * <code>structureData</code> must be in SMILES or molFile format.
     * </p>
     * <p>Example of Pageable:</p>
     * <code>
     * PageRequest pageable = new PageRequest(
     *           offset, limit, new Sort(
     *           new Order(Direction.ASC, "compoundName")));
     * </code>
     * <p>
     * The name given to Order ("compoundName") must equal a field name
     * of ChemicalCompound class.
     * </p>
     * <p>
     * A Predicate is a set of Expressions combined with boolean operator.<br>
     * A Predicate is created by using the querydsl generated query classes.
     * </p>
     * Example:<br>
     * <pre>
     * <code>
     * QChemicalCompound compound = QChemicalCompound.chemicalCompound;
     * QChemicalCompoundComposition composition =
     *           QChemicalCompoundComposition.chemicalCompoundComposition;
     *
     * Predicate predicate =
     *     compound.compoundName.like("Methyl%").and(composition.percentage.gt(50));
     *
     * </code>
     *</pre>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param structureData the chemical structure to search for
     * @param searchType FullStructure, SubStructure, SMARTS or Formula search
     * @param pageable contains offset, limit and sorting information. Must not be null.
     * @param predicate the filter (where clause) for determining which compounds to return.
     * @return a <code>Page</code> of compounds matching the given filters <code>structureData</code> and <code>predicate</code>.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate);

    /**
     * <p>
     * See
     * {@link #findByChemicalStructure(String, StructureSearchType,
     * Pageable, Predicate) findByChemicalStructure} for detailed information.
     * </p>
     * <p>
     * For possible values for
     * <code>searchOptions</code> see the
     * <a href="http://ggasoftware.com/opensource/bingo/user-manual-postgres#substructure-search">Bingo PostgreSQL cartridge documentation</a>.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param structureData the chemical structure to search for
     * @param searchType FullStructure, SubStructure, SMARTS or Formula search
     * @param pageable contains offset, limit and sorting information. Must not be null.
     * @param predicate the filter (where clause) for determining which compounds to return.
     * @param searchOptions for Bingo cartridge.
     * @return a <code>Page</code> of compounds matching the given filters <code>structureData</code> and <code>predicate</code>.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions);

    /**
     * <p>
     * See
     * {@link #findByChemicalStructure(String, StructureSearchType,
     * Pageable, Predicate) findByChemicalStructure} for detailed information.
     * </p>
     * <p>
     * For possible values for
     * <code>searchOptions</code> see the
     * <a href="http://ggasoftware.com/opensource/bingo/user-manual-postgres#substructure-search">Bingo PostgreSQL cartridge documentation</a>.
     * </p>
     *
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param structureData the chemical structure to search for
     * @param searchType FullStructure, SubStructure, SMARTS or Formula search
     * @param pageable contains offset, limit and sorting information. Must not be null.
     * @param predicate the filter (where clause) for determining which compounds to return.
     * @param searchOptions for Bingo cartridge.
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @return a <code>Page</code> of compounds matching the given filters <code>structureData</code> and <code>predicate</code>.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions, boolean loadCompositions, boolean loadContainables);

    /**
     * <p>Returns a set of ids for compounds matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType);

    /**
     * <p>Returns a set of ids for compounds matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate);

    /**
     * <p>Returns a set of ids for compounds matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            OrderSpecifier<?>... orders);

    /**
     * <p>Returns a set of ids for compounds matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            String searchOptions, OrderSpecifier<?>... orders);

    /**
     * <p>
     * Search for chemical structures similar to <code>structureData</code>
     * </p>
     * <p>
     * lowerBound and upperBound take values between 0 to 1 (including).
     * </p>
     * <p>
     * <code>structureData</code> must be in SMILES or molFile format.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param structureData the chemical structure to search for
     * @param similarityType the algorithm used for determining similarity
     * @param lowerBound    minimum required similarity
     * @param upperBound    maximum allowed similarity
     * @param pageable contains offset, limit and sorting information. Must not be null.
     * @return a <code>Page</code> of compounds matching the given filters <code>structureData</code> and <code>predicate</code>.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable);

    /**
     * <p>
     * Search for chemical structures similar to <code>structureData</code>
     * </p>
     * <p>
     * lowerBound and upperBound take values between 0 to 1 (including).
     * </p>
     * <p>
     * <code>structureData</code> must be in SMILES or molFile format.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param structureData the chemical structure to search for
     * @param similarityType the algorithm used for determining similarity
     * @param lowerBound    minimum required similarity
     * @param upperBound    maximum allowed similarity
     * @param pageable contains offset, limit and sorting information. Must not be null.
     * @param predicate     the filter (where clause) for determining which compounds to return.
     * @return a <code>Page</code> of compounds matching the given filters <code>structureData</code> and <code>predicate</code>.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate);

    /**
     * <p>
     * Search for chemical structures similar to <code>structureData</code>
     * </p>
     * <p>
     * lowerBound and upperBound take values between 0 to 1 (including).
     * </p>
     * <p>
     * <code>structureData</code> must be in SMILES or molFile format.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     *
     * @param structureData the chemical structure to search for
     * @param similarityType the algorithm used for determining similarity
     * @param lowerBound    minimum required similarity
     * @param upperBound    maximum allowed similarity
     * @param pageable contains offset, limit and sorting information. Must not be null.
     * @param predicate     the filter (where clause) for determining which compounds to return.
     * @param loadCompositions whether Lazy collection compositions should be loaded
     * @param loadContainables whether Lazy collection containables should be loaded
     * @return a <code>Page</code> of compounds matching the given filters <code>structureData</code> and <code>predicate</code>.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate,
            boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Counts number of compounds matching given predicate.
     * </p>
     * <p>
     * Predicate can be null. In that cause all records are counted.
     * </p>
     * <p>
     * If security is enabled, requires read role for this ChemicalCompound
     * implementation (eg. read_TestCompound)
     * </p>
     * @param predicate the filter (where clause) for determining which compounds to return.
     * @return number of compounds matching the given predicate
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    long count(Predicate predicate);

    /**
     * <p>
     * Creates or updates
     * <code>compound</code>. After calling this method it is crucial to
     * continue work with the returned compound (which can be a different
     * instance). So always call this method like below:
     * <pre>
     * <code>
     *
     * compound = chemicalCompoundService.save(compound);
     *
     * </code>
     * (google for JPA merge() vs. persist())
     * </pre>
     * </p>
     * <p>
     * If security is enabled, requires edit_created, create or write role for this ChemicalCompound
     * implementation (eg. create_TestCompound).
     * <p>
     * edit_created requires
     * </p>
     * <code>
     *      compound.getCreatedBy().equals(currentUserName) == true;
     * </code>
     * @param compound the compound to either create or update.
     * @return a compound which can be a different object than the input.
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasPermission(#compound, 'save_' + #root.this.getCompoundClassSimpleName())")
    @Override
    T save(T compound);

    /**
     * <p>
     * Imports an SD-File into the database. Each record in the sd-file is
     * considered to be a chemical compound and can contain disconnected
     * chemical structures.
     * </p>
     * <p>
     * No properties are imported with this method. The compoundName is set
     * to the compounds name in the sd-file. If that is null an exception will occur.
     * </p>
     * <p>
     * If you whish to load properties or percentages of different strutures
     * use an alternative <code>importSDF</code> method.
     * </p>
     * <p>
     * If <code>abortOnChemistryError</code> is true, Exceptions that occur due to
     * a chemical structure being invalid remain uncaught and transaction is
     * rolled back. If false, invalid structures are ignored and import
     * continues. However exceptions due to invalid property mappings, invalid
     * values,... are always thrown and wrapped in
     * <code>MoleculeDatabaseException</code>.
     * </p>
     * <p>
     * This is equivalent to calling {@link #importSDF(final Reader reader, final Map<String, String> mappedProperties,
            Map<String,Map<String,Double>> percentages, String recordIdentifier,
            * final boolean abortOnChemistryError) importSDF(reader, null, null, null, abortOnChemistryError)}.
     * </p>
     * <p>
     * If security is enabled, requires import role for this ChemicalCompound
     * implementation (eg. import_TestCompound)
     * </p>
     * @param reader the reader responsible for reading the SD-File
     * @param abortOnChemistryError whether to stop import if invalid structure data (molfile) is found
     * @return an <code>EntityImportResult</code> containing information about what compounds were imported.
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "compoundStructureQuery"}, allEntries = true)
    @PreAuthorize("hasRole('import_' + #root.this.getCompoundClassSimpleName())")
    EntityImportResult importSDF(final Reader reader, final boolean abortOnChemistryError);

    /**
     * <p>
     * Imports an SD-File into the database. Each record in the sd-file is
     * considered to be a chemical compound and can contain disconnected
     * chemical structures.
     * </p>
     * <p>
     * Properties are mapped by name (case-sensitive) were the key
     * is the property in the sd-file and the value is the field name of the
     * ChemicalCompound entity.<br> <br> If compoundName is not mapped, it isset
     * to the compounds name in the sd-file. If that is null an exception will occur.
     * <br> <br> Type is resolved by the fields type.
     * Only primitives, Strings or dates can be imported. Date / Calendar values
     * must be in one of the following formats: yyyy-MM-dd HH:mm:ss<br>
     * yyyy-MM-dd<br> dd.mm.yyyy<br> mm/dd/yyy<br> where HH is the hour value
     * from 0 to 23 (see java.text.SimpleDateFormat).
     * </p>
     * <p>
     * If you whish to load also load percentages of different strutures
     * use {@link #importSDF(final Reader reader, final Map<String, String> mappedProperties,
            Map<String,Map<String,Double>> percentages, String recordIdentifier,
            * final boolean abortOnChemistryError) this importSDF method}.
     * </p>
     * <p>
     * If <code>abortOnChemistryError</code> is true, Exceptions that occur due to
     * a chemical structure being invalid remain uncaught and transaction is
     * rolled back. If false, invalid structures are ignored and import
     * continues. However exceptions due to invalid property mappings, invalid
     * values,... are always thrown and wrapped in
     * <code>MoleculeDatabaseException</code>.
     * </p>
     * <p>
     * This is equivalent to calling {@link #importSDF(final Reader reader, final Map<String, String> mappedProperties,
            Map<String,Map<String,Double>> percentages, String recordIdentifier,
            * final boolean abortOnChemistryError) importSDF(reader, mappedProperties, null, null, abortOnChemistryError)}.
     * </p>
     * <p>
     * If security is enabled, requires import role for this ChemicalCompound
     * implementation (eg. import_TestCompound)
     * </p>
     * @param reader the reader responsible for reading the SD-File
     * @param mappedProperties maps SD-File properties to compound fields
     * @param abortOnChemistryError whether to stop import if invalid structure data (molfile) is found
     * @return an <code>EntityImportResult</code> containing information about what compounds were imported.
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "compoundStructureQuery"}, allEntries = true)
    @PreAuthorize("hasRole('import_' + #root.this.getCompoundClassSimpleName())")
    EntityImportResult importSDF(final Reader reader,
            final Map<String, String> mappedProperties,
            final boolean abortOnChemistryError);

    /**
     * <p>
     * Imports an SD-File into the database. Each record in the sd-file is
     * considered to be a chemical compound and can contain disconnected
     * chemical structures.
     * </p>
     * <p>
     * The percentage of a certain structure in a compound is mapped by defining
     * a unique property <code>recordIdentifier</code> in the sd-file. Then for each chemical
     * structure in a Compound create a mapping of the structure key to its
     * percentage in the compound. Add this Map to <code>percentages</code>
     * mapped by the unique property. If no mapping is found, percentages are set
     * to null. If the compound only has 1 structure, per default percentage is
     * set to 100 and no mapping is required.
     * </p>
     * <pre>
     * <code>
     *   String recordIdentifier = "Name";
     *   Map<String,Double> singleRecordPercentage = new HashMap<>();
     *   singleRecordPercentage.put("XDTMQSROBMDMFD-UHFFFAOYSA-N", 30.0);
     *   singleRecordPercentage.put("ZSWFCLXCOIISFI-UHFFFAOYSA-N", 70.0);
     *   Map<String,Map<String,Double>> percentages = new HashMap<>();
     *   percentages.put("Mixture_1", singleRecordPercentage);
     * </code>
     * </pre>
     * <p>
     * Properties are mapped by name (case-sensitive) were the key
     * is the property in the sd-file and the value is the field name of the
     * ChemicalCompound entity.<br> <br> If compoundName is not mapped, it isset
     * to the compounds name in the sd-file. If that is null an exception will occur.
     * <br> <br> Type is resolved by the fields type.
     * Only primitives, Strings or dates can be imported. Date / Calendar values
     * must be in one of the following formats: yyyy-MM-dd HH:mm:ss<br>
     * yyyy-MM-dd<br> dd.mm.yyyy<br> mm/dd/yyy<br> where HH is the hour value
     * from 0 to 23 (see java.text.SimpleDateFormat)
     * </p>
     * <p>
     * If <code>abortOnChemistryError</code> is true, Exceptions that occur due to
     * the chemical structure being invalid remain uncaught and transaction is
     * rolled back. If false, invalid structures are ignored and import
     * continues. However exceptions due to invalid property mappings, invalid
     * values,... are always wrapped in
     * <code>MoleculeDatabaseException</code>.
     * </p>
     * <p>
     * If security is enabled, requires import role for this ChemicalCompound
     * implementation (eg. import_TestCompound)
     * </p>
     * @param reader the reader responsible for reading the SD-File
     * @param mappedProperties maps SD-File properties to compound fields
     * @param percentages of all structures in case compound is a mixture
     * @param recordIdentifier uniquely identifies a compound in the given sd-file
     * @param abortOnChemistryError whether to stop import if invalid structure data (molfile) is found
     * @return an <code>EntityImportResult</code> containing information about what compounds were imported.
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "compoundStructureQuery"}, allEntries = true)
    @PreAuthorize("hasRole('import_' + #root.this.getCompoundClassSimpleName())")
    EntityImportResult importSDF(final Reader reader,
            final Map<String, String> mappedProperties,
            Map<String,Map<String,Double>> percentages,
            String recordIdentifier, final boolean abortOnChemistryError);

    /**
     * <p>
     * Export the given compounds to an SD-File without any properties.
     * </p>
     * <p>
     * If the compound contains of multiple structures, the percentages are
     * exported as a property of name structureKey and the value is the
     * percentage for that structure.
     * </p>
     * <p>
     * If security is enabled, requires export role for this ChemicalCompound
     * implementation (eg. export_TestCompound)
     * </p>
     *
     * @param compoundIds the ids of the compounds to export
     * @param writer the writer responsible for writing the output sd-file
     */
    @PreAuthorize("hasRole('export_' + #root.this.getCompoundClassSimpleName())")
    void exportSDF(Collection<Long> compoundIds, Writer writer);

    /**
     * <p>
     * Export the given compounds to an SD-File
     * </p>
     * <p>
     * If the compound contains of multiple structures, the percentages are
     * exported as a property of name InChikey and the value is the
     * percentage for that structure.
     * </p>
     * <p>
     *  The <code>Set<String> propertiesToExport</code> must contain the
     *  names of all the fields you want to export. If it is empty, only the structure
     *  will be exported, if it is null, all fields will be exported.
     * </p>
     * <p>
     * For date fields default format of "yyyy-MM-dd HH:mm:ss" is used.
     * </p>
     * <p>
     * If security is enabled, requires export role for this ChemicalCompound
     * implementation (eg. export_TestCompound)
     * </p>
     *
     * @param compoundIds the ids of the compounds to export
     * @param writer the writer responsible for writing the output sd-file
     */
    @PreAuthorize("hasRole('export_' + #root.this.getCompoundClassSimpleName())")
    void exportSDF(Collection<Long> compoundIds, Writer writer,
            Set<String> propertiesToExport);

    /**
     * <p>
     * Export the given compounds to an SD-File
     * </p>
     * <p>
     * If the compound contains of multiple structures, the percentages are
     * exported as a property of name InChikey and the value is the
     * percentage for that structure.
     * </p>
     * <p>
     *  The <code>Set<String> propertiesToExport</code> must contain the
     *  names of all the fields you want to export. If it is empty, only the structure
     *  will be exported, if it is null, all fields will be exported.
     * </p>
     * <p>
     * For date fields the date format defined by <code>dateFormat</code> is used.
     * See <a href="http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html">SimpleDateFormat</a>
     * for more information.
     * </p>
     * <p>
     * If security is enabled, requires export role for this ChemicalCompound
     * implementation (eg. export_TestCompound)
     * </p>
     *
     * @param compoundIds the ids of the compounds to export
     * @param writer the writer responsible for writing the output sd-file
     */
    @PreAuthorize("hasRole('export_' + #root.this.getCompoundClassSimpleName())")
    void exportSDF(Collection<Long> compoundIds, Writer writer,
            Set<String> propertiesToExport, String dateFormat);

    /**
     * <p>
     * Deletes the compound from the database
     * </p>
     * <p>
     * If security is enabled, requires delete or delete_created role for this ChemicalCompound
     * implementation (eg. delete_TestCompound or delete_created_TestCompound)
     * </p>
     * @param compound the compound to delete. Must not be null.
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "containableStructureQuery", "compoundStructureQuery"}, allEntries = true)
    @PreAuthorize("hasPermission(#compound, 'delete_' + #root.this.getCompoundClassSimpleName())")
    void delete(T compound);

    /**
     * <p>
     * Deletes the compound with the given id from the database.
     * </p>
     * <p>
     * If security is enabled, requires delete-Permission for this ChemicalCompound
     * implementation (eg. delete_TestCompound).
     * </p>
     * <p>
     * Due to a bug in Spring-Data this method will not work when using Spring
     * security (when using provided DefaultPermissionEvaluator). It will always
     * throw an AccessDenied exception.
     * </p>
     * @param compoundId the compound to delete. Must not be null.
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "containableStructureQuery","compoundStructureQuery"}, allEntries = true)
    @PreAuthorize("hasRole('delete_' + #root.this.getCompoundClassSimpleName())")
    void deleteById(Long compoundId);


    /**
     * <p>
     * Checks if <code>compound</code> violates any unique constraints.
     * </p>
     * <p>
     * If <code>returnValue.isUnique() == true</code>, the compound can be safely
     * inserted. If false, a call to save() will end in a
     * <code>DataIntegrityViolationException</code>.
     * </p>
     * <p>
     * The returned value contains a mapping of all fields an their values
     * that violate uniqueness.
     * </p>
     *
     * @param compound
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    @Override
    UniquenesscheckResult checkUniqueness(T compound);

    /**
     * <p>
     * Returns a compound create form the given input properties or null if it does not
     * exist in the database.
     * </p>
     * <p>
     * When importing compounds from SD-File this method checks if an equal
     * compound already exists in the database and if yes returns it. This should be
     * a "best-effort" to avoid unique constraint violation when inserting the
     * new imported entity.
     * </p>
     *
     * @param sdfRecord
     * @param mappedCompoundProperties
     * @return
     */
    @PreAuthorize("hasRole('import_' + #root.this.getCompoundClassSimpleName())")
    T getExistingCompound(SdfRecord sdfRecord,
            Map<String, String> mappedCompoundProperties);

    /**
     * <p>Renders a png image to <code>out</code>.
     *
     * <p>The image contains all the structures the compound is contained
     * of. The more structures there are the smaller each one will be.</p>
     *
     * @param compoundId the compounds unique database id
     * @param out the stream to which to ouput the image
     * @param width the total width of the image
     * @param height the total height of the image.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    public void renderChemicalCompound(Long compoundId, OutputStream out,
            int width, int height);

    /**
     * <p>Renders a png image to <code>out</code>.
     *
     * <p>The image contains all the structures the compound is contained
     * of.</p>
     *
     * <p>If fixed with is false, the width will be interpreted as the with that
     * each structure in the compound can use. So the total image width is
     * width * number of structures and the images width hence depends on the
     * specific compound and is not fixed.
     * </p>
     * @param compoundId the compounds unique database id
     * @param out the stream to which to ouput the image
     * @param width the width of the image but depends on fixedWidth
     * @param height the total height of the image.
     * @param fixedWidth if the image has a fixed width
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    public void renderChemicalCompound(Long compoundId, OutputStream out,
            int width, int height, boolean fixedWidth);

    /**
     * <p>Renders a png image to <code>out</code>.
     *
     * <p>The image contains all the structures the compound is contained
     * of and the part matching <code>highlightedSubstructure</code>
     * is highlighted.</p>
     *
     * <p>If fixed with is false, the width will be interpreted as the with that
     * each structure in the compound can use. So the total image width is
     * width * number of structures and the images width hence depends on the
     * specific compound and is not fixed.
     * </p>
     * @param compoundId the compounds unique database id
     * @param out the stream to which to ouput the image
     * @param width the width of the image but depends on fixedWidth
     * @param height the total height of the image.
     * @param fixedWidth if the image has a fixed width
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    public void renderChemicalCompound(Long compoundId, OutputStream out,
            int width, int height, boolean fixedWidth, String highlightedSubstructure);

    /**
     * Returns the PathBuilder used by this service.
     *
     * @return the PathBuilder used by this service.
     */
    PathBuilder<T> getPathBuilder();

    /**
     * return the Class of implementation of ChemicalCompound this Service is for.
     *
     * @return the Class of implementation of ChemicalCompound this Service is for.
     */
    @Override
    Class<T> getEntityClass ();

    /**
     * gets Simple Name of implementation of ChemicalCompound this Service is for.
     */
    String getCompoundClassSimpleName();

    /**
     * gets the name of the read role required to access search (read) methods
     * of this class.
     */
    String getReadRole();

}
