/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.google.common.base.Preconditions;
import com.mysema.query.types.Predicate;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.List;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalFormatConverter;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.bitbucket.kienerj.moleculedatabaseframework.io.IndigoMoleculeRenderer;
import org.bitbucket.kienerj.moleculedatabaseframework.repository.ChemicalStructureRepository;
import org.bitbucket.kienerj.moleculedatabaseframework.sql.EntityVersionConflictException;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.orm.ObjectOptimisticLockingFailureException;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p> Service class for data access of
 * <code>ChemicalStructure</code>. </p> <p> For saving
 * <code>ChemicalStructures</code> always use this service and never the
 * repository directly </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Service("chemicalStructureService")
@Transactional(readOnly = true)
public class ChemicalStructureServiceImpl<T extends ChemicalStructure>
        implements ChemicalStructureService<T> {

    private static final XLogger logger = XLoggerFactory.getXLogger("ChemicalStructureService");
    @Autowired
    @Qualifier("chemicalStructureRepository")
    private ChemicalStructureRepository<T> structureRepository;
    private ChemicalFormatConverter formatConverter = new ChemicalFormatConverter();

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getById(Long id) {
        logger.entry(id);
        Preconditions.checkNotNull(id);
        T structure = structureRepository.findOne(id);
        logger.exit(structure);
        return structure;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getByStructureKey(String structureKey) {
        logger.entry(structureKey);
        Preconditions.checkNotNull(structureKey);
        T structure = structureRepository.findByStructureKey(structureKey);
        logger.exit(structure);
        return structure;
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final T save(T structure) {
        logger.entry(structure);
        Preconditions.checkArgument(structure.getStructureKey() != null, "structureKey can not be null.");
        Preconditions.checkArgument(!structure.getStructureKey().equals(""), "structureKey can not be an empty String.");
        Preconditions.checkArgument(structure.getStructureData() != null, "structureData can not be null.");
        Preconditions.checkArgument(!structure.getStructureData().equals(""), "structureData can not be an empty String.");

        T result = structureRepository.findByStructureKey(structure.getStructureKey());

        if (result == null) {
            // clear id (and createdBy and created) so that a new row is inserted
            structure.reset();

            structure = structureRepository.save(structure);
            logger.debug("Creating new ChemicalStructure with structureKey {}.", structure.getStructureKey());
            logger.exit(structure);
            return structure;
        } else {
            logger.debug("Found exsiting ChemicalStructure with structureKey {}.", result.getStructureKey());
            logger.exit(result);
            return result;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final T updateExistingStructure(T structure) {

        logger.entry(structure);
        Preconditions.checkArgument(structure.getStructureKey() != null, "structureKey can not be null.");
        Preconditions.checkArgument(!structure.getStructureKey().equals(""), "structureKey can not be an empty String.");
        Preconditions.checkArgument(structure.getStructureData() != null, "structureData can not be null.");
        Preconditions.checkArgument(!structure.getStructureData().equals(""), "structureData can not be an empty String.");
        Preconditions.checkArgument(structure.getId() != null, "ChemicalStructure must have an id.");

        T result = structureRepository.findOne(structure.getId());

        if (result == null) {
            IllegalArgumentException ex = new IllegalArgumentException("Given ChemicalStructure does not exist. It can't be updated.");
            logger.throwing(ex);
            throw ex;
        }
        try {
            logger.debug("Updating ChemicalStructure with structureKey {}.", structure.getStructureKey());
            structure = structureRepository.save(structure);
            logger.exit(structure);
            return structure;
        } catch (ObjectOptimisticLockingFailureException lockEx) {
            if (lockEx.getPersistentClassName().equals(structure.getClass().getName())) {
                T currentVersion = getById(structure.getId());
                throw new EntityVersionConflictException(lockEx, currentVersion, structure);
            } else {
                throw lockEx;
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final long count(Predicate predicate) {
        logger.entry(predicate);
        long count = structureRepository.count(predicate);
        logger.exit(count);
        return count;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<ChemicalCompound> getOccurences(ChemicalStructure structure) {
        logger.entry(structure);
        Preconditions.checkNotNull(structure);
        List<ChemicalCompound> result = structureRepository.getChemicalCompoundsByStructure(structure);
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<ChemicalCompoundComposition> getCompositions(ChemicalStructure structure) {
        logger.entry(structure);
        Preconditions.checkNotNull(structure);
        List<ChemicalCompoundComposition> result =
                structureRepository.getCompositionsByStructure(structure);
        loadCompounds(result);
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<ChemicalCompound> getOccurences(Long structureId) {
        logger.entry(structureId);
        Preconditions.checkNotNull(structureId);
        List<ChemicalCompound> result = structureRepository.getChemicalCompoundsByStructureId(structureId);
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<ChemicalCompoundComposition> getCompositions(Long structureId) {
        logger.entry(structureId);
        Preconditions.checkNotNull(structureId);
        List<ChemicalCompoundComposition> result = structureRepository.getCompositionsByStructureId(structureId);
        loadCompounds(result);
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public UniquenesscheckResult checkUniqueness(T structure){
        Preconditions.checkNotNull(structure);
        Preconditions.checkNotNull(structure.getStructureKey());
        T result = getByStructureKey(structure.getStructureKey());

        HashMap<String, Object> fields = new HashMap<>();
        boolean isUnique = true;

        if (result != null){
            fields.put("structureKey", structure.getStructureKey());
            isUnique = false;
        }
        return new UniquenesscheckResult(isUnique, fields);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void renderChemicalStructure(Long structureId, OutputStream out,
            int width, int height) {

        IndigoMoleculeRenderer renderer = new IndigoMoleculeRenderer(width, height);
        renderer.renderChemicalStructure(getById(structureId).getStructureData(), out);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void renderChemicalStructure(Long structureId, OutputStream out,
            int width, int height, String highlightedSubstructure){

        IndigoMoleculeRenderer renderer = new IndigoMoleculeRenderer(width, height);
        renderer.renderChemicalStructure(getById(structureId).getStructureData(),
                out, highlightedSubstructure);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final String getStandardInChi(String structureData) {
        logger.entry(structureData);
        Preconditions.checkNotNull(structureData);
        String inchi = formatConverter.toStandardInchi(structureData);
        logger.exit(inchi);
        return inchi;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final String getStandardInChiKey(String structureData) {
        logger.entry(structureData);
        Preconditions.checkNotNull(structureData);
        String inchi = formatConverter.toStandardInchikey(structureData);
        logger.exit(inchi);
        return inchi;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final String getInChi(String structureData, String options) {
        logger.entry(structureData, options);
        Preconditions.checkNotNull(structureData);
        options = options == null ? "" : options;
        String inchi = formatConverter.toInchi(structureData, options);
        logger.exit(inchi);
        return inchi;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final String getInChiKey(String structureData, String options) {
        logger.entry(structureData, options);
        Preconditions.checkNotNull(structureData);
        options = options == null ? "" : options;
        String inchi = formatConverter.toInchi(structureData, options);
        String inchiKey = formatConverter.inchiToInchikey(inchi);
        logger.exit(inchiKey);
        return inchiKey;
    }

    @Override
    public Class getEntityClass(){
        return ChemicalStructure.class;
    }

    private void loadCompounds(List<ChemicalCompoundComposition> compositions) {
        if (compositions == null) {
            return;
        }
        // load relationship
        for (ChemicalCompoundComposition composition : compositions) {
            composition.getCompound().getId();
        }
    }
}
