/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import java.util.List;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Transactional(readOnly=true)
public interface CompositionService {

    /**
     * Get a composition by its numeric unique id
     */
    @PreAuthorize("hasRole('read_Composition')")
    ChemicalCompoundComposition getById(Long id);

    /**
     * Gets all compositions for the given compound.
     */
    @PreAuthorize("hasRole('read_Composition')")
    List<ChemicalCompoundComposition> getByCompoundId(Long compoundId);

    /**
     * Gets all compositions using the given chemical structure.
     */
    @PreAuthorize("hasRole('read_Composition')")
    List<ChemicalCompoundComposition> getByStructureId(Long structureId);

    /**
     * <p>Adds or updated the given <code>composition</code>.</p>
     *
     * <p>
     * <code>composition</code> must have a <code>ChemicalStructure</code> and
     * <code>ChemicalCompound</code> set (eg. != null) and the used
     * <code>ChemicalCompound</code> must be attached to the session or else an
     * <code>InvalidDataAccessApiUsageException</code> will be thrown.
     * </p>
     *
     * <p>
     * The composition must be saved before it is added to the compound
     * <pre>
     * <ocde>
     * composition = compositionService.save(composition);
     * compound.addCompostion(composition);
     * </code>
     * <pre>
     * or a NullPointerException can be thrown.
     * </p>
     * @param composition
     * @return
     * @throws InvalidDataAccessApiUsageException if <code>composition.getCompound()</code> returns a detached entity.
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasRole('save_Composition')")
    ChemicalCompoundComposition save(ChemicalCompoundComposition composition);

    /**
     * Deletes the according composition from the database and removes it
     * from the associated compounds set of compositions
     * @param compositionId
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "containableStructureQuery","compoundStructureQuery"}, allEntries = true)
    @PreAuthorize("hasRole('delete_Composition')")
    void deleteById(Long compositionId);

    /**
     * Deletes the according composition from the database and removes it
     * from the associated compounds set of compositions
     * @param composition
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "containableStructureQuery","compoundStructureQuery"}, allEntries = true)
    @PreAuthorize("hasRole('delete_Composition')")
    void delete(ChemicalCompoundComposition composition);
}
