/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.google.common.base.Preconditions;
import java.util.List;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.bitbucket.kienerj.moleculedatabaseframework.repository.ChemicalCompoundCompositionRepository;
import org.bitbucket.kienerj.moleculedatabaseframework.sql.EntityVersionConflictException;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

@Service("compositionService")
@Transactional(readOnly = true)
public class CompositionServiceImpl implements CompositionService {

    private static final XLogger logger = XLoggerFactory.getXLogger("CompositionService");
    private static final String OPT_LOCK_EX_MESSAGE = "Could not update composition "
            + "with id %d due to a concurrent modification. Call getCurrentVersion()"
            + "for an updated version of the conflicting entity.";
    @Autowired
    @Qualifier("compositionRepository")
    private ChemicalCompoundCompositionRepository repository;
    @Autowired
    private ChemicalStructureService chemicalStructureService;

    /**
     * {@inheritDoc}
     */
    @Override
    public ChemicalCompoundComposition getById(Long id) {
        Preconditions.checkNotNull(id);
        ChemicalCompoundComposition result = repository.findByCompositionId(id);
        result.getCompound().getId(); // initialize proxy
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<ChemicalCompoundComposition> getByCompoundId(Long compoundId) {
        Preconditions.checkNotNull(compoundId);
        List<ChemicalCompoundComposition> result = repository.findByPkCompoundId(compoundId);
        for (ChemicalCompoundComposition composition : result) {
            composition.getCompound().getId(); // initialize proxy
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<ChemicalCompoundComposition> getByStructureId(Long structureId) {
        Preconditions.checkNotNull(structureId);
        List<ChemicalCompoundComposition> result = repository.findByPkChemicalStructureId(structureId);
        for (ChemicalCompoundComposition composition : result) {
            composition.getCompound().getId(); // initialize proxy
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public ChemicalCompoundComposition save(ChemicalCompoundComposition composition) {
        Preconditions.checkNotNull(composition);
        Preconditions.checkNotNull(composition.getChemicalStructure());
        Preconditions.checkNotNull(composition.getCompound());
        Preconditions.checkArgument(composition.getCompound().getId() != null, "ChemicalCompound must be an existing compound.");

        ChemicalStructure structure = composition.getChemicalStructure();
        ChemicalStructure actualStructure = chemicalStructureService.save(structure);
        composition.setChemicalStructure(actualStructure);

        if (composition.getId() != null) {
            // inserts chemical structure if it is new
            // required because else next statement will fail if it is new
            repository.flush();
            // this is an update
            // use native query to update else there are ton of annyoing
            // issues and exception caused by JPA (hibernate)
            // JPA IS F RETARDED as you can't update something with a composite
            // primary key only delete and create a new one.
            int updateCount = repository.updateCompoundComposition(composition.getId(),
                composition.getChemicalStructure().getId(),
                composition.getCompound().getId(),
                composition.getPercentage(),
                composition.getVersion());
            // if updateCount is 0, no record was updated and hence a
            // concurrent modification occured.
            if (updateCount == 0) {
                ChemicalCompoundComposition currentVersion = getById(composition.getId());
                throw new EntityVersionConflictException(
                        String.format(OPT_LOCK_EX_MESSAGE, composition.getId()),
                        currentVersion, composition);
            }
            ChemicalCompoundComposition result = getById(composition.getId());
            result.getCompound().getId(); // initialize proxy
            return result;
//            ChemicalCompoundComposition old = getById(composition.getId());
//            composition.getCompound().getCompositions().remove(old);
//            repository.delete(old);
//            repository.flush();
//            ChemicalCompoundComposition result = repository.saveAndFlush(composition);
//            return result;
        } else {
            // flush to get Composition ID generated in database
            ChemicalCompoundComposition result = repository.saveAndFlush(composition);
            result.getCompound().addComposition(result);
            repository.flush();
            return result;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public void deleteById(Long compositionId) {
        Preconditions.checkNotNull(compositionId);
        // get entity by id and delete it.
        delete(getById(compositionId));
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public void delete(ChemicalCompoundComposition composition) {
        Preconditions.checkNotNull(composition);
        repository.delete(composition);
        // It is required to explicitly remove this compostion
        // from the Compounds Set of compostions.
        composition.getCompound().getCompositions().remove(composition);
    }
}
