/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.path.PathBuilder;
import java.io.Reader;
import java.io.Writer;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.bitbucket.kienerj.moleculedatabaseframework.io.EntityImportResult;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.SimilarityType;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p>
 * A <code>CompoundContainerService</code> manages data access for
 * <code>ChemicalCompoundContainer</code> entities. This includes methods for saving,
 * retrieving and searching for <code>ChemicalCompoundContainers</code>.
 * </p>
 * <p>
 * Methods returning a <code>ChemicalCompoundContainer</code> that do not have
 * parameters <code>boolean loadCompositions</code> and <code> boolean loadContainables</code>
 * will load all <code>ChemicalCompoundCompositions</code> of the associated <code>ChemicalCompound</code>.
 * However <code>Containables</code> are not loaded hence </p>
 * <code>result.getChemicalCompound().getContainables();</code>
 * <p>
 * will cause a LazyInitializationException outside of a entityManager session.
 * </p>
 * <p>
 * All service methods are transactional and annotated with Spring Security. Security
 * must be explictly enabled in configuration.
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Transactional(readOnly=true)
public interface CompoundContainerService<T extends ChemicalCompoundContainer>
        extends Service<T> {

    /**
     * <p>
     * Returns a list of all <code>ChemicalCompoundContainers</code> the <code>Containable</code>
     * with id <code>containableId</code> is contained in.
     * </p>
     *
     * @param containableId
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> getByContainable(Long containableId);

    /**
     * Gets a container by its internal id
     *
     * @param id
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T getById(final Long id);

    /**
     * <p>
     * Gets a container by its internal id
     * </p>
     * <p>
     * loadCompositions and loadContainables determine if the according Lazy collections
     * in the associated <code>ChemicalCompound</code> should be loaded or not.
     * </p>
     * <p>
     * In case of false <code>getChemicalCompound().getCompostions()</code> or
     * <code>getChemicalCompound().getContainables()</code> will cause a
     * LazyInitializationException outside of a EntityManager session.
     * </p>
     *
     * @param id
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T getById(final Long id, boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Gets a Container by its barcode
     * </p>
     *
     * @param barcode
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T getByBarcode(String barcode);

    /**
     * <p>
     * Gets a Container by its barcode
     * </p>
     *
     * <p>
     * loadCompositions and loadContainables determine if the according Lazy collections
     * in the associated ChemicalCompound should be loaded or not.
     * </p>
     * <p>
     * In case of false <code>getChemicalCompound().getCompostions()</code> or
     * <code>getChemicalCompound().getContainables()</code> will cause a
     * LazyInitializationException outside of a EntityManager session.
     * </p>
     *
     * @param barcode
     * @param loadCompositions
     * @param loadContainables
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T getByBarcode(String barcode, boolean loadCompositions, boolean loadContainables);

    /**
     * Gets all containers with an id that is contained in
     * <code>ids</code>.
     *
     * @param ids
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> getById(final List<Long> ids);

    /**
     * <p>
     * Gets all containers with an id that is contained in
     * <code>ids</code>.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * @param ids
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> getById(final List<Long> ids, boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Returns an unsorted List of all containers id matching the predicate.
     * </p>     *
     * <p>
     * <code>predicate</code> must not be null or an NullPointertException is thrown.
     * </p>     *
     * <p>
     * If security is enabled, requires read role for this Container
     * implementation (eg. read_TestContainer)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which containers to return. must not be null.
     * @return a <code>List</code> of all containers ids matching the given <code>predicate</code>
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> getAllIds(Predicate predicate);


    /**
     * Returns 1 container matching the given predicate.
     *
     * @param predicate
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T findOne(Predicate predicate);

    /**
     * <p>
     * Returns 1 container matching the given predicate.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * @param predicate
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T findOne(Predicate predicate, boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Returns an unsorted List of all containers matching the predicate.
     * </p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * </p>
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate);

    /**
     * <p>
     * Returns an unsorted List of all containers matching the predicate.
     * </p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate, boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Returns a sorted List of all containers matching the predicate.
     * </p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * <p>
     * @param predicate
     * @param orders
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate, OrderSpecifier<?>... orders);

    /**
     * <p>
     * Returns a sorted List of all containers matching the predicate.
     * </p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * <p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * @param predicate
     * @param orders
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate, boolean loadCompositions,
            boolean loadContainables, OrderSpecifier<?>... orders);

    /**
     * Returns a Page of containers matching the given predicate.
     *
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findAll(Predicate predicate, Pageable pageable);

    /**
     * <p>
     * Returns a Page of containers matching the given predicate.
     *</p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findAll(Predicate predicate, Pageable pageable,
            boolean loadCompositions, boolean loadContainables);

    /**
     * Find containers containing a compound with CAS Number <code>cas</code>.
     *
     * @param cas
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findByCas(String cas);

    /**
     * <p>
     * Find containers containing a compound with CAS Number <code>cas</code>.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * @param cas
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findByCas(String cas, boolean loadCompositions, boolean loadContainables);

//    /**
//     * <p>
//     * Returns a List of containers that contain a compound which consists
//     * of <code>chemicalStructure</code>. Because ChemicalStructures are unique
//     * this is equivalent to a full structure search.
//     * </p>
//     * @param structure
//     * @return
//     */
//    @PreAuthorize("hasRole(#root.this.getReadRole())")
//    @PostFilter("hasPermission(filterObject, 'read_' + T(org.springframework.util.ClassUtils).getShortName(filterObject.getChemicalCompound().class))")
//    List<T> findByChemicalStructureEntity(ChemicalStructure structure);
//
//    /**
//     * <p>
//     * Returns a List of containers that contain a compound which consists
//     * of <code>chemicalStructure</code>. Because ChemicalStructures are unique
//     * this is equivalent to a full structure search.
//     * </p>
//     * <p>
//     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
//     * for further information on parameters loadCompositions and loadContainables.
//     * </p>
//     * @param structure
//     * @return
//     */
//    @PreAuthorize("hasRole(#root.this.getReadRole())")
//    @PostFilter("hasPermission(filterObject, 'read_' + T(org.springframework.util.ClassUtils).getShortName(filterObject.getChemicalCompound().class))")
//    List<T> findByChemicalStructureEntity(ChemicalStructure structure,
//            boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Returns a List of containers that contain <code>compound</code>.
     * </p>
     * @param compound
     * @return
     */
    @PreAuthorize("hasRole('read_' + #root.this.getContainerClassSimpleName()) and "
            + "hasRole('read_' + T(org.springframework.util.ClassUtils)"
            + ".getShortName(#compound.class))")
    List<T> findByChemicalCompound(ChemicalCompound compound);

    /**
     * <p>
     * Returns a List of containers that contain <code>compound</code>.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     * @param compound
     * @return
     */
    @PreAuthorize("hasRole('read_' + #root.this.getContainerClassSimpleName()) and "
            + "hasRole('read_' + T(org.springframework.util.ClassUtils)"
            + ".getShortName(#compound.class))")
    List<T> findByChemicalCompound(ChemicalCompound compound,
            boolean loadCompositions, boolean loadContainables);

    /**
     * <p>
     * Returns a Page of containers that contain at least 1 chemical structure
     * matching the chemical structure search arguments.
     * </p>
     * <p>
     * Example of Pageable:
     * </p>
     * <pre>
     * <code>
     * PageRequest pageable = new PageRequest(
     *           offset, limit, new Sort(
     *           new Order(Direction.ASC, "containerName")));
     * </code>
     * </pre>
     * <p>
     * The name given to Order ("containerName") must equal a field name
     * of ChemicalCompoundContainer class.
     * </p>
     *
     * @param structureData
     * @param searchType
     * @param pageable
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable);

    /**
     * <p>
     * Returns a Page of containers that contain at least 1 chemical structure
     * matching the chemical structure search arguments and the predicate.
     * </p>
     * <p>
     * Example of Pageable:
     * </p>
     * <pre>
     * <code>
     * PageRequest pageable = new PageRequest(
     *           offset, limit, new Sort(
     *           new Order(Direction.ASC, "containerName")));
     * </code>
     * </pre>
     * <p>
     * The name given to Order ("containerName") must equal a field name
     * of ChemicalCompoundContainer class.
     * </p>
     *
     * <p>A Predicate is a set of Expressions combined with boolean operator.</p>
     * <p>A Predicate is created by using the querydsl generated query classes.</p>
     *
     * <p>Example:</p>
     * <pre>
     * <code>
     *
     * QBasicCompoundContainer basicContainer =
                QBasicCompoundContainer.basicCompoundContainer;
        QChemicalCompound compound = QChemicalCompound.chemicalCompound;
        QChemicalCompoundComposition composition =
                QChemicalCompoundComposition.chemicalCompoundComposition;
     *
     * Predicate predicate =
     *     container.containerName.like("Methyl%").and(composition.percentage.gt(50));
     *
     * </code>
     * </pre>
     * @param structureData
     * @param searchType
     * @param pageable
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate);

    /**
     * <p>
     * See
     * {@link #findByChemicalStructure(String, StructureSearchType,
     * Pageable, Predicate) findByChemicalStructure} for detailed information.
     *</p>
     * <p>
     * For possible values for
     * <code>searchOptions</code> see the
     * <a href="http://ggasoftware.com/opensource/bingo/user-manual-postgres#substructure-search">Bingo PostgreSQL cartridge documentation</a>.
     * </p>
     *
     * @param structureData
     * @param searchType
     * @param pageable
     * @param predicate
     * @param searchOptions
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions);

    /**
     * <p>
     * See
     * {@link #findByChemicalStructure(String, StructureSearchType,
     * Pageable, Predicate) findByChemicalStructure} for detailed information.
     * </p>
     * <p>
     * For possible values for
     * <code>searchOptions</code> see the
     * <a href="http://ggasoftware.com/opensource/bingo/user-manual-postgres#substructure-search">Bingo PostgreSQL cartridge documentation</a>.
     * </p>
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * @param structureData
     * @param searchType
     * @param pageable
     * @param predicate
     * @param searchOptions
     *
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions, boolean loadCompositions, boolean loadContainables);


    /**
     * <p>Returns a list of ids for containers matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType);

    /**
     * <p>Returns a list of ids for containers matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate);

    /**
     * <p>Returns a list of ids for containers matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            OrderSpecifier<?>... orders);

    /**
     * <p>Returns a list of ids for containers matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            String searchOptions, OrderSpecifier<?>... orders);

    /**
     * <p>
     * Search for chemical structures similar to <code>structureData</code>
     * </p>
     * <p>lowerBound and upperBound take values between 0 to 1 (including).</p>
     *
     * @param structureData
     * @param similarityType the algorithm used for determining similarity
     * @param lowerBound    minimum required similarity
     * @param upperBound    maximum allowed similarity
     * @param pageable      for sorting and paging
     * @param predicate     other criteria beside similarity
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate);


    /**
     * <p>
     * Search for chemical structures similar to <code>structureData</code>
     * </p>
     * <p>lowerBound and upperBound take values between 0 to 1 (including).</p>
     *
     * <p>
     * See {@link #getById(Long id, boolean loadCompositions, boolean loadContainables) getById}
     * for further information on parameters loadCompositions and loadContainables.
     * </p>
     *
     * @param structureData
     * @param similarityType the algorithm used for determining similarity
     * @param lowerBound    minimum required similarity
     * @param upperBound    maximum allowed similarity
     * @param pageable      for sorting and paging
     * @param predicate     other criteria beside similarity
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate,
            boolean loadCompositions, boolean loadContainables);

    /**
     * Counts number of containers matching given predicate.
     *
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    long count(Predicate predicate);

    /**
     * <p>Creates or updates <code>container</code>.</p>
     * <p>
     * After calling this method it is crucial to
     * continue work with the returned container (which can be a different
     * instance). So always call this method like below:
     * </p>
     * <pre>
     * <code>
     * container = compoundContainerService.save(container);     *
     * </code>
     * </pre>
     * <p>(google for JPA merge() vs. persist())</p>
     *
     * @param container
     * @return
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasPermission(#container, 'save_' + #root.this.getContainerClassSimpleName())")
    @Override
    T save(T container);



    /**
     * * <p>
     * Imports an SD-File into the database. Each record in the sd-file is
     * considered to be a chemical compound container.
     * </p>
     * <p>
     * Properties are mapped to the according entity  (Container, Containable, Compound)
     * by the according Maps <code>mappedCompoundProperties, mappedcontainableProperties,
     * mappedContainerProperties</code>.<br>
     * The method also requires a unique property for Containable. This is used
     * to check if the containable already exists and if yes is reused. This is
     * very important. <b>In case a containable already exists, it is reused including
     * the associated ChemicalCompound and any structural information of the
     * current SD-File entry are completely ignored!!!</b>
     * </p>
     * <p>
     * compoundClass and containableClass determine the type of instances
     * created for Compound and Containable. A Container import can only import
     * 1 type of ChemicalCompound and Containable.
     * </p>
     * <p>
     * The Compound in the container can contain disconnected chemical structures.
     * The percentage of a certain structure in a compound is mapped by defining
     * a unique property <code>compoundIdentifier</code> in the sd-file. Then for each chemical
     * structure in a Compound create a mapping of the InChIkey to its
     * percentage in the compound. Add this Map to <code>percentages</code>
     * mapped by the unique property. If no mapping is found, percentages are set
     * to null. If the compound only has 1 structure, per default percentage is
     * set to 100 and no mapping is required.
     * </p>
     * <pre>
     * <code>
     *   String compoundIdentifier = "Name";
     *   Map<String,Double> singleRecordPercentage = new HashMap<>();
     *   singleRecordPercentage.put("XDTMQSROBMDMFD-UHFFFAOYSA-N", 30.0);
     *   singleRecordPercentage.put("ZSWFCLXCOIISFI-UHFFFAOYSA-N", 70.0);
     *   Map<String,Map<String,Double>> percentages = new HashMap<>();
     *   percentages.put("Mixture_1", singleRecordPercentage);
     * </code>
     * </pre>
     * <p>
     * Properties are mapped by name (case-sensitive) were the key
     * is the property in the sd-file and the value is the field name of the
     * entity.<br> <br> If compoundName is not mapped, it is set
     * to the compounds name in the sd-file. If that is null an exception will occur.
     * <br> <br> Type is resolved by the fields type.
     * Only primitives, Strings or dates can be imported. Date / Calendar values
     * must be in one of the following formats: yyyy-MM-dd HH:mm:ss<br>
     * yyyy-MM-dd<br> dd.mm.yyyy<br> mm/dd/yyy<br> where HH is the hour value
     * from 0 to 23 (see java.text.SimpleDateFormat)
     * </p>
     * <p>
     * If <code>abortOnChemistryError</code> is true, Exceptions that occur due to
     * the chemical structure being invalid remain uncaught and transaction is
     * rolled back. If false, invalid structures are ignored and import
     * continues. However exceptions due to invalid property mappings, invalid
     * values,... are always wrapped in
     * <code>MoleculeDatabaseException</code>.
     * </p>
     *
     *
     * @param <R> ChemicalCompound Type
     * @param <S> Containable Type
     * @param reader
     * @param mappedCompoundProperties
     * @param compoundClass
     * @param mappedContainableProperties
     * @param containableClass
     * @param uniqueContainableProperty
     * @param mappedContainerProperties
     * @param percentages
     * @param compoundIdentifier
     * @param abortOnChemistryError
     * @return
     */
    <R extends ChemicalCompound, S extends Containable> EntityImportResult importSDF(final Reader reader,
            final Map<String, String> mappedCompoundProperties,
            final Class<R> compoundClass,
            final Map<String, String> mappedContainableProperties,
            final Class<S> containableClass,
            final String uniqueContainableProperty,
            final Map<String, String> mappedContainerProperties,
            final Map<String, Map<String, Double>> percentages,
            final String compoundIdentifier, final boolean abortOnChemistryError);


    /**
     * <p>
     * Exports containers to an SD-File.
     * </p>
     * <p>
     * If the container contains of multiple structures, the percentages are
     * exported as a property of name InChIkey and the value is the
     * percentage for that structure.
     * </p>
     * <p>
     *  The <code>Sets<String></code> must contain the
     *  names of all the fields you want to export for the according entity.
     *  If it is empty, only the structure will be exported, if it is null, all
     * fields will be exported.
     * </p>
     * <p>
     * For date fields default format of "yyyy-MM-dd HH:mm:ss" is used.
     * </p>
     * <p>
     * If security is enabled, requires export role for this ChemicalCompoundContainer
     * implementation (eg. export_TestCompoundContainer)
     * </p>
     *
     * @param containerIds the ids of the containers to export
     * @param writer the writer responsible for writing the output sd-file
     */
    @PreAuthorize("hasRole('export_' + #root.this.getContainerClassSimpleName())")
    void exportSDF(List<Long> containerIds, Writer writer,
            Set<String> containerProperties, Set<String> containableProperties,
            Set<String> compoundProperties);


    /**
     * <p>
     * Exports containers to an SD-File.
     * </p>
     * <p>
     * If the container contains of multiple structures, the percentages are
     * exported as a property of name InChIkey and the value is the
     * percentage for that structure.
     * </p>
     * <p>
     *  The <code>Sets<String></code> must contain the
     *  names of all the fields you want to export for the according entity.
     *  If it is empty, only the structure will be exported, if it is null, all
     * fields will be exported.
     * </p>
     * <p>
     * For date fields the date format defined by <code>dateFormat</code> is used.
     * See <a href="http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html">SimpleDateFormat</a>
     * for more information.
     * </p>
     * <p>
     * If security is enabled, requires export role for this ChemicalCompoundContainer
     * implementation (eg. export_TestCompoundContainer)
     * </p>
     *
     * @param containerIds the ids of the containers to export
     * @param writer the writer responsible for writing the output sd-file
     */
    @PreAuthorize("hasRole('export_' + #root.this.getContainerClassSimpleName())")
    void exportSDF(List<Long> containerIds, Writer writer,
            Set<String> containerProperties, Set<String> containableProperties,
            Set<String> compoundProperties, String dateFormat);


    /**
     * Deletes the container from the database
     *
     * @param container
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasPermission(#container, 'delete_' + #root.this.getContainerClassSimpleName())")
    void delete(T container);

    /**
     * Deletes the container with the given id from the database
     *
     * @param containerId
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasRole('delete_' + #root.this.getContainerClassSimpleName())")
    void deleteById(Long containerId);

    /**
     * <p>
     * Checks if <code>container</code> already exists in the database by looking
     * up all unique constrained fields.
     * </p>
     * <p>
     * If <code>returnValue.isUnique() == true</code>, the container can be safely
     * inserted. If false, a call to save() will end in a
     * <code>DataIntegrityViolationException</code>.
     * </p>
     * <p>
     * The returned value contains a mapping of all fields an their values
     * that violate uniqueness.
     * </p>     *
     * @param containable
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    @Override
    UniquenesscheckResult checkUniqueness(T container);


    PathBuilder<T> getPathBuilder();
}
