/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.google.common.base.Preconditions;
import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.path.PathBuilder;
import java.io.Reader;
import java.io.Writer;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import javax.annotation.Resource;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;
import org.bitbucket.kienerj.moleculedatabaseframework.MoleculeDatabaseFrameworkException;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalCompoundFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalStructureFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.QChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.io.ContainableFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.io.ContainerFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.io.EntityImportResult;
import org.bitbucket.kienerj.moleculedatabaseframework.io.RawSdfReader;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfContainerExporter;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfContainerCreator;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfRecord;
import org.bitbucket.kienerj.moleculedatabaseframework.repository.CompoundContainerRepository;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.SimilarityType;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p> Service Class for
 * <code>ChemicalCompoundContainer</code> entity. </p> <p> Methods returning a
 * <code>ChemicalCompoundContainer</code> that do not have parameters
 * <code>boolean loadCompositions</code> and
 * <code> boolean loadContainables</code> will load all
 * <code>ChemicalCompoundCompositions</code> of the associated
 * <code>ChemicalCompound</code>. However
 * <code>Containables</code> are not loaded hence </p>
 * <code>result.getChemicalCompound().getContainables();</code> <p> will cause a
 * LazyInitializationException outside of a entityManager session. </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Resource // must be Resource and not Service due to Maps in constructor not having a string key!!!
@Transactional(readOnly = true)
public class CompoundContainerServiceImpl<T extends ChemicalCompoundContainer<? extends Containable>>
        implements CompoundContainerService<T> {

    private static final XLogger logger = XLoggerFactory.getXLogger("CompoundContainerService");
    private static final boolean LOAD_COMPOSITIONS = true;
    private static final boolean LOAD_CONTAINABLES = false;
    private static final int NUMBER_OF_IMPORT_THREADS = Runtime.getRuntime().availableProcessors();
    @PersistenceContext
    private EntityManager entityManager;
    @Autowired
    @Getter
    @Setter(AccessLevel.PROTECTED)
    private CompoundContainerRepository<T> compoundContainerRepository;
    @Autowired
    private ChemicalStructureService chemicalStructureService;
    @Autowired
    private ChemicalStructureFactory chemicalStructureFactory;
    @Getter
    private Class<T> entityClass;
    @Getter
    private PathBuilder<T> pathBuilder;
    // required for Spring Security SPEL expressions because
    // service.getClass().getSimpleName() is not valid SPEL
    @Getter
    private String containerClassSimpleName;
    @Getter
    private final String readRole;
    private Map<Class<? extends ChemicalCompound>, ChemicalCompoundService<? extends ChemicalCompound>> compoundServices;
    private Map<Class<? extends Containable>, ContainableService<? extends Containable>> containableServices;

    @Autowired
    public CompoundContainerServiceImpl(Class<T> compoundContainerClass,
            Map<Class<? extends ChemicalCompound>, ChemicalCompoundService<? extends ChemicalCompound>> compoundServices,
            Map<Class<? extends Containable>, ContainableService<? extends Containable>> containableServices) {

        logger.entry(compoundContainerClass, compoundServices, containableServices);

        Preconditions.checkNotNull(compoundContainerClass);
        Preconditions.checkNotNull(compoundServices);
        Preconditions.checkNotNull(containableServices);
//        Preconditions.checkArgument(!compoundServices.isEmpty());
//        Preconditions.checkArgument(!containableServices.isEmpty());

        this.entityClass = compoundContainerClass;
        this.containerClassSimpleName = compoundContainerClass.getSimpleName();
        this.readRole = "read_" + containerClassSimpleName;
        this.pathBuilder = new PathBuilder<>(entityClass, "chemicalCompoundContainer");
        this.compoundServices = compoundServices;
        this.containableServices = containableServices;
        logger.exit();
    }

    protected CompoundContainerRepository<T> getRepository() {
        return this.compoundContainerRepository;
    }

    /**
     * <p> This method is called in the save() method before the entity
     * <code>container</code> is actually passed down to the JPA provider for
     * actual persisting (or merging). In case the entity requires some
     * operations or checks to be done before saving it, do it in this method.
     * </p> <p> Override this method if you need to. Per default it does nothing
     * at all. </p> <p> An example is setting a non-nullable value to a database
     * generated value if the value was not supplied by the user. </p>
     *
     */
    protected void preSave(T container) {
    }

    /**
     * <p> Called in save() method immediately after the entity was saved. </p>
     * <p> This can be used as example to force a flush after every save. Per
     * default flush only happens at end of transaction. </p>
     */
    protected void postSave(T container) {
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> getByContainable(Long containableId) {
        logger.entry(containableId);
        Preconditions.checkNotNull(containableId);
        Predicate predicate = QChemicalCompoundContainer.chemicalCompoundContainer.containable.id.eq(containableId);
        predicate = addSecurityFilter(predicate);
        List<T> containers = toList(getCompoundContainerRepository().findAll(predicate), LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getById(Long id) {
        logger.entry(id);
        T container = getById(id, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(container);
        return container;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getById(Long id, boolean loadCompositions, boolean loadContainables) {
        logger.entry(id, loadCompositions, loadContainables);
        Preconditions.checkNotNull(id);
        Predicate predicate = QChemicalCompoundContainer.chemicalCompoundContainer.id.eq(id);
        predicate = addSecurityFilter(predicate);
        T container = getRepository().findOne(predicate);
        loadLazyCompoundCollections(container, loadCompositions, loadContainables);
        logger.exit(container);
        return container;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getByBarcode(String barcode) {
        logger.entry(barcode);
        T container = getByBarcode(barcode, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(container);
        return container;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getByBarcode(String barcode, boolean loadCompositions, boolean loadContainables) {
        logger.entry(barcode, loadCompositions, loadContainables);
        Preconditions.checkNotNull(barcode);
        Predicate predicate = QChemicalCompoundContainer.chemicalCompoundContainer.barcode.eq(barcode);
        predicate = addSecurityFilter(predicate);
        T container = getRepository().findOne(predicate);
        loadLazyCompoundCollections(container, loadCompositions, loadContainables);
        logger.exit(container);
        return container;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> getById(List<Long> ids) {
        logger.entry(ids);
        List<T> containers = getById(ids, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> getById(List<Long> ids, boolean loadCompositions, boolean loadContainables) {
        logger.entry(ids, loadCompositions, loadContainables);
        Preconditions.checkNotNull(ids);
        Predicate predicate = QChemicalCompoundContainer.chemicalCompoundContainer.id.in(ids);
        predicate = addSecurityFilter(predicate);
        List<T> containers = toList(compoundContainerRepository.findAll(predicate),
                loadCompositions, loadContainables);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> getAllIds(Predicate predicate) {
        logger.entry(predicate);
        List<Long> ids = getRepository().findAll(predicate, pathBuilder);
        logger.exit(ids);
        return ids;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T findOne(Predicate predicate) {
        logger.entry(predicate);
        T container = findOne(predicate, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(container);
        return container;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T findOne(Predicate predicate, boolean loadCompositions, boolean loadContainables) {
        logger.entry(predicate, loadCompositions, loadContainables);
        Preconditions.checkNotNull(predicate);
        predicate = addSecurityFilter(predicate);
        T container = getRepository().findOne(predicate);
        loadLazyCompoundCollections(container, loadCompositions, loadContainables);
        logger.exit(container);
        return container;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate) {
        logger.entry(predicate);
        List<T> containers = findAll(predicate, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate, boolean loadCompositions, boolean loadContainables) {
        logger.entry(predicate);
        Preconditions.checkNotNull(predicate);
        predicate = addSecurityFilter(predicate);
        List<T> containers = toList(getRepository().findAll(predicate), loadCompositions, loadContainables);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate, OrderSpecifier<?>... orders) {
        logger.entry(predicate, orders);
        List<T> containers = findAll(predicate, LOAD_COMPOSITIONS, LOAD_CONTAINABLES, orders);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate, boolean loadCompositions,
            boolean loadContainables, OrderSpecifier<?>... orders) {
        logger.entry(predicate, orders);
        Preconditions.checkNotNull(predicate);
        predicate = addSecurityFilter(predicate);
        List<T> containers = toList(getRepository().findAll(predicate, orders),
                loadCompositions, loadContainables);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findAll(Predicate predicate, Pageable pageable) {
        logger.entry(predicate, pageable);
        Page<T> page = findAll(predicate, pageable, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findAll(Predicate predicate, Pageable pageable,
            boolean loadCompositions, boolean loadContainables) {
        logger.entry(predicate, pageable);
        Preconditions.checkNotNull(predicate);
        Preconditions.checkNotNull(pageable);
        predicate = addSecurityFilter(predicate);
        Page<T> page = getRepository().findAll(predicate, pageable);
        loadLazyCompoundCollections(page, loadCompositions, loadContainables);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findByCas(String cas) {
        logger.entry(cas);
        List<T> containers = findByCas(cas, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findByCas(String cas, boolean loadCompositions, boolean loadContainables) {
        logger.entry(cas);
        Preconditions.checkNotNull(cas);
        Predicate predicate = QChemicalCompoundContainer.chemicalCompoundContainer.containable.chemicalCompound.cas.eq(cas);
        predicate = addSecurityFilter(predicate);
        List<T> containers = toList(compoundContainerRepository.findAll(predicate), loadCompositions, loadContainables);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findByChemicalCompound(ChemicalCompound compound) {
        logger.entry(compound);
        List<T> containers = findByChemicalCompound(compound, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findByChemicalCompound(ChemicalCompound compound,
            boolean loadCompositions, boolean loadContainables) {
        logger.entry(compound);
        Preconditions.checkNotNull(compound);
        List<T> containers = compoundContainerRepository.findByContainableChemicalCompound(compound);
        loadLazyCompoundCollections(containers, loadCompositions, loadContainables);
        logger.exit(containers);
        return containers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable) {

        logger.entry(structureData, searchType, pageable);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        Predicate predicate = addSecurityFilter(null);
        Page<T> page = getRepository().findByChemicalStructure(
                structureData, searchType, pageable, predicate, "",
                pathBuilder);
        loadLazyCompoundCollections(page, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable, Predicate predicate) {

        logger.entry(structureData, searchType, pageable);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        predicate = addSecurityFilter(predicate);
        Page<T> page = getRepository().findByChemicalStructure(
                structureData, searchType, pageable, predicate, "",
                pathBuilder);
        loadLazyCompoundCollections(page, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable, Predicate predicate,
            String searchOptions) {

        logger.entry(structureData, searchType, pageable);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        predicate = addSecurityFilter(predicate);
        Page<T> page = getRepository().findByChemicalStructure(
                structureData, searchType, pageable, predicate, searchOptions,
                pathBuilder);
        loadLazyCompoundCollections(page, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable, Predicate predicate,
            String searchOptions, boolean loadCompositions, boolean loadContainables) {

        logger.entry(structureData, searchType, pageable);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        predicate = addSecurityFilter(predicate);
        Page<T> page = getRepository().findByChemicalStructure(
                structureData, searchType, pageable, predicate, searchOptions,
                pathBuilder);
        loadLazyCompoundCollections(page, loadCompositions, loadContainables);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType) {

        return findByChemicalStructure(structureData, searchType, null, "");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate) {

        return findByChemicalStructure(structureData, searchType, predicate, "");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            OrderSpecifier<?>... orders) {

        return findByChemicalStructure(structureData, searchType, predicate, "", orders);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            String searchOptions, OrderSpecifier<?>... orders) {

        logger.entry(structureData, searchType);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        predicate = addSecurityFilter(predicate);
        List<Long> ids = getRepository().findByChemicalStructure(structureData,
                searchType, predicate, searchOptions, pathBuilder, orders);
        logger.exit(ids);
        return ids;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate) {

        logger.entry(structureData, similarityType, lowerBound, upperBound, pageable, predicate);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(similarityType);
        Preconditions.checkNotNull(lowerBound); // upperBound can be null (=default of 1.0)
        Preconditions.checkNotNull(pageable);
        predicate = addSecurityFilter(predicate);
        Page<T> page = getRepository().findBySimilarStructure(structureData, similarityType,
                lowerBound, upperBound, pageable, predicate, pathBuilder);
        loadLazyCompoundCollections(page, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate,
            boolean loadCompositions, boolean loadContainables) {

        logger.entry(structureData, similarityType, lowerBound, upperBound, pageable, predicate);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(similarityType);
        Preconditions.checkNotNull(lowerBound); // upperBound can be null (=default of 1.0)
        Preconditions.checkNotNull(pageable);
        predicate = addSecurityFilter(predicate);
        Page<T> page = getRepository().findBySimilarStructure(structureData, similarityType,
                lowerBound, upperBound, pageable, predicate, pathBuilder);
        loadLazyCompoundCollections(page, loadCompositions, loadContainables);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final long count(Predicate predicate) {
        logger.entry(predicate);
        predicate = addSecurityFilter(predicate);
        long count = getRepository().count(predicate);
        logger.exit(count);
        return count;
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    @CacheEvict(value = "containerStructureQuery", allEntries = true)
    public final T save(T container) {
        logger.entry(container);
        Preconditions.checkNotNull(container);
        preSave(container);
        // if entity container.getContainable already exists, it
        // must first be reattached to the entity manager or else
        // an exception will occur (issue in Spring Data JPA ->
        // save() method internal calls persists intead of merge)
        if (container.getId() == null
                && container.getContainable().getId() != null) {
            Containable containable = container.getContainable();
            ContainableService containableService = containableServices
                    .get(containable.getClass());
            containable = containableService.getById(containable.getId());
            container.setContainable(containable.getClass().cast(containable));
        }
        // for new entites, set ChemicalStructures correctly because
        // of cascading of persist. Merge is not cascaded hence do nothing.
        boolean isNewCompound = container.getChemicalCompound().getId() == null;
        if (isNewCompound
                && container.getChemicalCompound().getCompositions() != null) {
            for (ChemicalCompoundComposition composition : container.getChemicalCompound().getCompositions()) {
                ChemicalStructure structure = composition.getChemicalStructure();
                ChemicalStructure result = chemicalStructureService.save(structure);
                composition.setChemicalStructure(result);
            }
        }

        container = getRepository().save(container);
        postSave(container);
        logger.exit(container);
        return container;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final <R extends ChemicalCompound, S extends Containable> EntityImportResult importSDF(final Reader reader,
            final Map<String, String> mappedCompoundProperties,
            final Class<R> compoundClass,
            final Map<String, String> mappedContainableProperties,
            final Class<S> containableClass,
            final String uniqueContainableProperty,
            final Map<String, String> mappedContainerProperties,
            final Map<String, Map<String, Double>> percentages,
            final String compoundIdentifier, final boolean abortOnChemistryError) {

        logger.entry(reader, mappedCompoundProperties, compoundClass,
                mappedContainableProperties, containableClass,
                mappedContainerProperties, percentages, compoundIdentifier);
        Preconditions.checkNotNull(reader);

        final RawSdfReader sdfReader = new RawSdfReader(reader);

        final ChemicalCompoundFactory<R> compoundFactory =
                new ChemicalCompoundFactory<>(chemicalStructureFactory,
                compoundClass);
        final ContainableFactory<S> containableFactory = new ContainableFactory(containableClass);
        final ContainerFactory<T> containerFactory = new ContainerFactory(entityClass);


        final ChemicalCompoundService<R> compoundService = getCompoundService(compoundClass);
        final ContainableService<S> containableService = getContainableService(containableClass);

        final EntityImportResult importResult = new EntityImportResult();

        int numWorkerThreads = 2; // more do not increase performance
        int maxQueueSize = 1000;

        final ExecutorService taskGenerator = Executors.newFixedThreadPool(1);
        final ExecutorService taskExecutor = Executors.newFixedThreadPool(numWorkerThreads);
        final LinkedBlockingQueue<Future<T>> futures = new LinkedBlockingQueue<>(maxQueueSize);

        Runnable generator = new Runnable() {
            @Override
            public void run() {
                while (sdfReader.hasNext()) {

                    SdfRecord sdfRecord = sdfReader.next();

                    // don't import if container with same barcode already exists
                    if (mappedContainerProperties != null && mappedContainerProperties.containsValue("barcode")) {
                        String barcodeKey = null;
                        for (Map.Entry<String, String> entry : mappedContainerProperties.entrySet()) {
                            if (entry.getValue().equals("barcode")) {
                                barcodeKey = entry.getKey();
                                break;
                            }
                        }
                        String barcode = sdfRecord.getProperty(barcodeKey);
                        T container = getByBarcode(barcode);
                        if (container != null) {
                            importResult.addAlreadyPresentEntity(container.getId());
                            continue;
                        }
                    }

                    Map<String, Double> currentPercentages = null;
                    if (compoundIdentifier != null && percentages != null) {
                        String identifier = sdfRecord.getProperties()
                                .get(compoundIdentifier);
                        currentPercentages = percentages.get(identifier);
                        logger.debug("Found percentages for compound {}.", identifier);
                    }

                    // if compound or containable already exists, use that compound and / or containable
                    ChemicalCompound compound = compoundService
                            .getExistingCompound(sdfRecord, mappedCompoundProperties);
                    Containable containable = containableService
                            .getExistingContainable(sdfRecord, mappedContainableProperties, mappedCompoundProperties);

                    // pass compound and containable to ContainerCreator
                    // if they are null, a new instance will be created in the importer
                    // This is done here that creator does not need to access the database
                    // and could be run in a separate thread.
                    SdfContainerCreator<T> sdfContainerCreator =
                            new SdfContainerCreator<>(sdfRecord, compoundFactory,
                            containableFactory, containerFactory,
                            mappedCompoundProperties, mappedContainableProperties,
                            mappedContainerProperties, currentPercentages,
                            compound, containable,
                            importResult, abortOnChemistryError);

                    Future<T> containerFuture =
                            taskExecutor.submit(sdfContainerCreator);
                    try {
                        futures.put(containerFuture);
                    } catch (InterruptedException ex) {
                        Thread.currentThread().interrupt();
                        throw new MoleculeDatabaseFrameworkException(ex);
                    }
                }
                taskExecutor.shutdown();
            }
        };
        try {
            Future<?> generatorResult = taskGenerator.submit(generator);

            int counter = 0;
            while (!generatorResult.isDone()
                    || !futures.isEmpty()) {
                Future<T> containerFuture = futures.take();
                T container = containerFuture.get();
                // due CascadeType.PERSIST -> also saves containable, Compound...
                container = save(container);
                importResult.addImportedEntity(container.getId());
                counter++;
                // Set hibernate.jdbc.batch_size to 100 too
                // massive performance increase with large inserts
                // at least 10 times faster
                if (counter >= 100) {
                    counter = 0;
                    entityManager.flush();
                    entityManager.clear();
                    logger.debug("Batch Inserting Containers: entityManager.flush() and entityManager.clear() called.");
                }
            }
            // if not called all exceptions thrown in task generator
            // will be "lost" and method would complete successfully.
            generatorResult.get();
            logger.exit(importResult);
            return importResult;
        } catch (InterruptedException ex) {
            Thread.currentThread().interrupt();
            throw new MoleculeDatabaseFrameworkException(ex);
        } catch (ExecutionException ex) {
            throw new MoleculeDatabaseFrameworkException(ex);
        } finally {
            taskExecutor.shutdownNow();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void exportSDF(List<Long> containerIds, Writer writer,
            Set<String> containerProperties, Set<String> containableProperties,
            Set<String> compoundProperties) {

        logger.entry(containerIds, writer, containerProperties, containableProperties,
                compoundProperties);
        Preconditions.checkNotNull(writer);
        Preconditions.checkNotNull(containerIds);

        SdfContainerExporter<T> exporter = new SdfContainerExporter<>(this);

        exporter.exportContainersToSdf(containerIds, writer, containerProperties,
                containableProperties, compoundProperties);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void exportSDF(List<Long> containerIds, Writer writer,
            Set<String> containerProperties, Set<String> containableProperties,
            Set<String> compoundProperties, String dateFormat) {

        logger.entry(containerIds, writer, containerProperties, containableProperties,
                compoundProperties, dateFormat);
        Preconditions.checkNotNull(writer);
        Preconditions.checkNotNull(containerIds);
        Preconditions.checkNotNull(dateFormat);

        SdfContainerExporter<T> exporter = new SdfContainerExporter<>(this);

        exporter.exportContainersToSdf(containerIds, writer, containerProperties,
                containableProperties, compoundProperties, dateFormat);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    @CacheEvict(value = "containerStructureQuery", allEntries = true)
    public final void delete(T container) {
        logger.entry(container);
        Preconditions.checkNotNull(container);
        getRepository().delete(container);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    @CacheEvict(value = "containerStructureQuery", allEntries = true)
    public final void deleteById(Long containerId) {
        logger.entry(containerId);
        Preconditions.checkNotNull(containerId);
        getRepository().delete(containerId);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public UniquenesscheckResult checkUniqueness(T container) {
        QChemicalCompoundContainer qContainer = new QChemicalCompoundContainer("chemicalCompoundContainer");
        Predicate predicate = qContainer.barcode.eq(container.getBarcode());
        T result = getRepository().findOne(predicate);

        HashMap<String, Object> fields = new HashMap<>();
        boolean isUnique = true;

        if (result != null) {
            fields.put("cas", container.getBarcode());
            isUnique = false;
        }

        return new UniquenesscheckResult(isUnique, fields);
    }

    /**
     * Returns the service for the given compound class.
     *
     * @param <R>
     * @param compoundClass
     * @return
     */
    protected final <R extends ChemicalCompound> ChemicalCompoundService<R> getCompoundService(Class<R> compoundClass) {
        ChemicalCompoundService<R> compoundService = (ChemicalCompoundService<R>) compoundServices.get(compoundClass);
        if (compoundService == null) {
            throw new IllegalArgumentException(String.format("Class %s has no service "
                    + "associated to it in CompoundContainerService.",
                    compoundClass.getSimpleName()));
        }
        return compoundService;
    }

    /**
     * Returns the service for the given containable class.
     *
     * @param <S>
     * @param containableClass
     * @return
     */
    protected final <S extends Containable> ContainableService<S> getContainableService(Class<S> containableClass) {

        ContainableService<S> containableService = (ContainableService<S>) containableServices.get(containableClass);
        if (containableService == null) {
            throw new IllegalArgumentException(String.format("Class %s has no service "
                    + "associated to it in CompoundContainerService.",
                    containableClass.getSimpleName()));
        }
        return containableService;
    }

    /**
     * <p> If you use spring security any custom search method must only use
     * <code>findOne(predicate)</code> and
     * <code>findAll(predicate)</code> repository methods.<br> Example from
     * getByBarcode() method: </p>
     * <pre>
     * <code>
     * Predicate predicate = QChemicalCompoundContainer.chemicalCompoundContainer.barcode.eq(barcode);
     * predicate = addSecurityFilter(predicate);
     * </code>
     * </pre> <p> This enforces that a user never sees a container containing
     * compounds he does nto have the read privilege for. </p>
     *
     *
     * @param predicate the predictae to which to add the security filter
     * @return
     */
    protected final Predicate addSecurityFilter(Predicate predicate) {
        if (SecurityContextHolder.getContext().getAuthentication() != null) {
            String userName = SecurityContextHolder.getContext().getAuthentication().getName();
            QChemicalCompoundContainer qContainer = QChemicalCompoundContainer.chemicalCompoundContainer;
            predicate = qContainer.containable.chemicalCompound.readRole.users.any().username.eq(userName)
                    .and(qContainer.containable.readRole.users.any().username.eq(userName))
                    .and(predicate);
        }
        return predicate;
    }

    protected final List<T> toList(Iterable<T> iterable, boolean loadCompositions, boolean loadContainables) {
        if (iterable == null) {
            return null;
        }
        List<T> containers = new ArrayList<>();
        for (T container : iterable) {
            containers.add(container);
            loadLazyCompoundCollections(container, loadCompositions, loadContainables);
        }
        return containers;
    }

    protected final void loadLazyCompoundCollections(Iterable<T> page, boolean loadCompositions, boolean loadContainables) {
        if (page != null && (loadCompositions || loadContainables)) {
            for (T container : page) {
                loadLazyCompoundCollections(container, loadCompositions, loadContainables);
            }
        }
    }

    protected final void loadLazyCompoundCollections(T container, boolean loadCompositions, boolean loadContainables) {
        if (container != null) {
            if (loadCompositions) {
                for (ChemicalCompoundComposition composition : container.getChemicalCompound().getCompositions()) {
                    composition.getChemicalStructure().getId();
                }
            }
            if (loadContainables) {
                for (Containable containable : container.getChemicalCompound().getContainables()) {
                    containable.getId();
                }
            }
        }
    }
}
