/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.google.common.base.Preconditions;
import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.path.PathBuilder;
import java.beans.Introspector;
import java.io.Reader;
import java.io.Writer;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import lombok.Getter;
import org.bitbucket.kienerj.moleculedatabaseframework.MoleculeDatabaseFrameworkException;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalCompoundFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalStructureFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.QContainable;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Role;
import org.bitbucket.kienerj.moleculedatabaseframework.io.ContainableFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.io.EntityImportResult;
import org.bitbucket.kienerj.moleculedatabaseframework.io.RawSdfReader;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfContainableExporter;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfContainableCreator;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfRecord;
import org.bitbucket.kienerj.moleculedatabaseframework.repository.ContainableRepository;
import org.bitbucket.kienerj.moleculedatabaseframework.repository.RoleRepository;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.SimilarityType;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p> Service class for data access of any type of
 * <code>Containable</code>. </p> <p> To create a service for your
 * <code>Containable</code> implementations, your need to extend
 * <code>ContainableService</code> and create an implementation extending this
 * class and implementing your service interface. </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Service
@Transactional(readOnly = true)
public abstract class ContainableServiceImpl<T extends Containable, R extends ChemicalCompound> implements ContainableService<T> {

    private static final XLogger logger = XLoggerFactory.getXLogger("ContainableService");
    private static final int NUMBER_OF_IMPORT_THREADS = Runtime.getRuntime().availableProcessors();
    private static final boolean LOAD_COMPOSITIONS = true;
    @PersistenceContext
    private EntityManager entityManager;
    @Autowired
    private ChemicalStructureService chemicalStructureService;
    @Autowired
    private RoleRepository roleRepository;
    @Autowired
    private ChemicalStructureFactory chemicalStructureFactory;
    @Getter
    private final Class<T> entityClass;
    @Getter
    private final Class<R> compoundClass;
    @Getter
    private final PathBuilder<T> pathBuilder;
    @Getter
    private final String containableClassSimpleName;
    @Getter
    private final String readRole;

    @Autowired
    public ContainableServiceImpl(Class<T> containableClass, Class<R> compoundClass) {
        Preconditions.checkNotNull(containableClass);
        Preconditions.checkNotNull(compoundClass);
        this.entityClass = containableClass;
        this.compoundClass = compoundClass;
        this.containableClassSimpleName = containableClass.getSimpleName();
        this.readRole = "read_" + containableClassSimpleName;
        String path = Introspector.decapitalize(containableClass.getSimpleName());
        pathBuilder = new PathBuilder<>(containableClass, path);
    }

    /**
     * Returns the Spring Data Repository used for all database access
     *
     */
    protected abstract ContainableRepository<T> getRepository();

    /**
     * <p> Gets the ChemicalCompoundService for the ChemicalCompound associated
     * with this Containable of Type T. </p> <p> As example for
     * <code>TestContainable extends Containable<TestCompound></code> this
     * method will return a
     * <code>ChemicalCompoundService<TestCompound></code>. </p>
     */
    protected abstract ChemicalCompoundService<R> getChemicalCompoundService();

    /**
     * <p> This method is called in the save() method before the entity
     * <code>containable</code> is actually passed down to the JPA provider for
     * actual persisting (or merging). In case the entity requires some
     * operations or checks to be done before saving it, do it in this method.
     * </p> <p> Override this method if you need to. Per default it does nothing
     * at all. </p> <p> An example is setting a non-nullable value to a database
     * generated value if the value was not supplied by the user. </p>
     *
     */
    protected void preSave(T containable) {
    }

    /**
     * <p> Called in save() method immediately after the entity was saved. </p>
     * <p> This can be used as example to force a flush after every save. Per
     * default flush only happens at end of transaction. </p>
     */
    protected void postSave(T containable) {
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public abstract T getExistingContainable(SdfRecord sdfRecord,
            Map<String, String> mappedContainableProperties,
            Map<String, String> mappedCompoundProperties);

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getById(Long id) {
        logger.entry(id);
        Preconditions.checkNotNull(id);
        T containable = getRepository().findOne(id);
        loadCompositions(containable, LOAD_COMPOSITIONS);
        logger.exit(containable);
        return containable;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> getById(List<Long> ids) {
        logger.entry(ids);
        Preconditions.checkNotNull(ids);
        List<T> containables = toList(getRepository().findAll(ids), LOAD_COMPOSITIONS);
        logger.exit(containables);
        return containables;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> getByChemicalCompound(ChemicalCompound<T> compound) {
        logger.entry(compound);
        Preconditions.checkNotNull(compound);
        Preconditions.checkNotNull(compound.getId());
        List<T> containables = getByChemicalCompoundId(compound.getId());
        logger.exit(containables);
        return containables;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> getByChemicalCompoundId(Long compoundId) {
        logger.entry(compoundId);
        Preconditions.checkNotNull(compoundId);
        QContainable containable =
                new QContainable(pathBuilder);
        Predicate predicate = containable.chemicalCompound.id.eq(compoundId);
        List<T> containables = toList(getRepository().findAll(predicate), LOAD_COMPOSITIONS);
        logger.exit(containables);
        return containables;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> getAllIds(Predicate predicate) {
        logger.entry(predicate);
        List<Long> ids = getRepository().findAll(predicate, pathBuilder);
        logger.exit(ids);
        return ids;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T findOne(Predicate predicate) {
        logger.entry(predicate);
        Preconditions.checkNotNull(predicate);
        T containable = getRepository().findOne(predicate);
        loadCompositions(containable, LOAD_COMPOSITIONS);
        logger.exit(containable);
        return containable;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate) {
        logger.entry(predicate);
        Preconditions.checkNotNull(predicate);
        List<T> containables = toList(getRepository().findAll(predicate), LOAD_COMPOSITIONS);
        logger.exit(containables);
        return containables;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate, OrderSpecifier<?>... orders) {
        logger.entry(predicate, orders);
        Preconditions.checkNotNull(predicate);
        List<T> containables = toList(getRepository().findAll(predicate, orders), LOAD_COMPOSITIONS);
        logger.exit(containables);
        return containables;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findAll(Predicate predicate, Pageable pageable) {
        logger.entry(predicate, pageable);
        Preconditions.checkNotNull(predicate);
        Preconditions.checkNotNull(pageable);
        Page<T> containables = getRepository().findAll(predicate, pageable);
        loadCompositions(containables, LOAD_COMPOSITIONS);
        logger.exit(containables);
        return containables;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable) {

        logger.entry(structureData, searchType, pageable);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        Page<T> page = getRepository().findByChemicalStructure(
                structureData, searchType, pageable, null, "",
                pathBuilder);
        loadCompositions(page, LOAD_COMPOSITIONS);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable, Predicate predicate) {

        logger.entry(structureData, searchType, pageable);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        Page<T> page = getRepository().findByChemicalStructure(
                structureData, searchType, pageable, predicate, "",
                pathBuilder);
        loadCompositions(page, LOAD_COMPOSITIONS);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable, Predicate predicate,
            String searchOptions) {

        logger.entry(structureData, searchType, pageable);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        Page<T> page = getRepository().findByChemicalStructure(
                structureData, searchType, pageable, predicate, searchOptions,
                pathBuilder);
        loadCompositions(page, LOAD_COMPOSITIONS);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable, Predicate predicate,
            String searchOptions, boolean loadCompositions) {

        logger.entry(structureData, searchType, pageable);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        Page<T> page = getRepository().findByChemicalStructure(
                structureData, searchType, pageable, predicate, searchOptions,
                pathBuilder);
        loadCompositions(page, LOAD_COMPOSITIONS);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType) {

        return findByChemicalStructure(structureData, searchType, null, "");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate) {

        return findByChemicalStructure(structureData, searchType, predicate, "");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            OrderSpecifier<?>... orders) {

        return findByChemicalStructure(structureData, searchType, predicate, "", orders);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            String searchOptions, OrderSpecifier<?>... orders) {

        logger.entry(structureData, searchType);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);

        List<Long> ids = getRepository().findByChemicalStructure(structureData,
                searchType, predicate, searchOptions, pathBuilder, orders);
        logger.exit(ids);
        return ids;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate) {

        logger.entry(structureData, similarityType, lowerBound, upperBound, pageable, predicate);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(similarityType);
        Preconditions.checkNotNull(lowerBound); // upperBound can be null (=default of 1.0)
        Preconditions.checkNotNull(pageable);
        Page<T> page = getRepository().findBySimilarStructure(structureData, similarityType,
                lowerBound, upperBound, pageable, predicate, pathBuilder);
        loadCompositions(page, LOAD_COMPOSITIONS);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate,
            boolean loadCompositions) {

        logger.entry(structureData, similarityType, lowerBound, upperBound, pageable, predicate);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(similarityType);
        Preconditions.checkNotNull(lowerBound); // upperBound can be null (=default of 1.0)
        Preconditions.checkNotNull(pageable);
        Page<T> page = getRepository().findBySimilarStructure(structureData, similarityType,
                lowerBound, upperBound, pageable, predicate, pathBuilder);
        loadCompositions(page, loadCompositions);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final long count(Predicate predicate) {
        logger.entry(predicate);
        long count = getRepository().count(predicate);
        logger.exit(count);
        return count;
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "containableStructureQuery"}, allEntries = true)
    @Override
    public final T save(T containable) {
        logger.entry(containable);
        Preconditions.checkNotNull(containable);
        preSave(containable);
        // if entity containable.getChemicalCompound already exists, it
        // must first be reattached to the entity manager or else
        // an exception will occur (issue in Spring Data JPA ->
        // save() method internal calls persists intead of merge)
        if (containable.getId() == null
                && containable.getChemicalCompound().getId() != null) {
            ChemicalCompound compound = getChemicalCompoundService()
                    .getById(containable.getChemicalCompound().getId());
            containable.setChemicalCompound(compound);
        }
        // for new entites, set ChemicalStructures correctly because
        // of cascading of persist. Merge is not cascaded hence do nothing.
        boolean isNewCompound = containable.getChemicalCompound().getId() == null;
        if (isNewCompound
                && containable.getChemicalCompound().getCompositions() != null) {
            Set<ChemicalCompoundComposition> compositions = containable.getChemicalCompound().getCompositions();
            for (ChemicalCompoundComposition composition : compositions) {
                ChemicalStructure structure = composition.getChemicalStructure();
                ChemicalStructure result = chemicalStructureService.save(structure);
                composition.setChemicalStructure(result);
            }
        }
        if (containable.getReadRole() == null) {
            Role role = roleRepository.findByRoleName("read_" + getContainableClassSimpleName());
            if (role == null) {
                role = new Role("read_" + getContainableClassSimpleName());
                role = roleRepository.save(role);
            }
            containable.setReadRole(role);
        }
        containable = getRepository().save(containable);
        postSave(containable);
        logger.exit(containable);
        return containable;
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final EntityImportResult importSDF(final Reader reader,
            final Map<String, String> mappedCompoundProperties,
            final Map<String, String> mappedContainableProperties,
            final Map<String, Map<String, Double>> percentages,
            final String compoundIdentifier, final boolean abortOnChemistryError) {

        logger.entry(reader, mappedCompoundProperties, compoundClass,
                mappedContainableProperties, entityClass, percentages,
                compoundIdentifier);
        Preconditions.checkNotNull(reader);

        final RawSdfReader sdfReader = new RawSdfReader(reader);

        final ChemicalCompoundFactory<R> compoundFactory =
                new ChemicalCompoundFactory<>(chemicalStructureFactory,
                compoundClass);
        final ContainableFactory<T> containableFactory = new ContainableFactory(entityClass);

        final EntityImportResult importResult = new EntityImportResult();
        int numWorkerThreads = 2; // more do not increase performance
        int maxQueueSize = 1000;

        final ExecutorService taskGenerator = Executors.newFixedThreadPool(1);
        final ExecutorService taskExecutor = Executors.newFixedThreadPool(numWorkerThreads);
        final LinkedBlockingQueue<Future<T>> futures = new LinkedBlockingQueue<>(maxQueueSize);

        try {

            Runnable generator = new Runnable() {
                @Override
                public void run() {

                    while (sdfReader.hasNext()) {

                        SdfRecord sdfRecord = sdfReader.next();

                        T existingContainable = getExistingContainable(
                                sdfRecord, mappedContainableProperties, mappedCompoundProperties);

                        if (existingContainable != null) {
                            importResult.addAlreadyPresentEntity(existingContainable.getId());
                            continue;
                        }

                        Map<String, Double> currentPercentages = null;
                        if (compoundIdentifier != null && percentages != null) {
                            String identifier = sdfRecord.getProperties().get(compoundIdentifier);
                            currentPercentages = percentages.get(identifier);
                            logger.debug("Found percentages for compound {}.", identifier);
                        }

                        // if compound already exists, use that compound
                        R compound = getChemicalCompoundService().getExistingCompound(sdfRecord, mappedCompoundProperties);

                        // pass compound to containable creator
                        // if compound is null, a new instance will be created in the creator
                        // This is done here that creator does not need to access the database
                        // and could be run in a separate thread.
                        SdfContainableCreator<T> sdfContainableCreator =
                                new SdfContainableCreator<>(sdfRecord, compoundFactory,
                                containableFactory, mappedCompoundProperties, mappedContainableProperties,
                                currentPercentages,
                                compound, importResult, abortOnChemistryError);
                        Future<T> compoundFuture =
                                taskExecutor.submit(sdfContainableCreator);
                        try {
                            futures.put(compoundFuture);
                        } catch (InterruptedException ex) {
                            Thread.currentThread().interrupt();
                            throw new MoleculeDatabaseFrameworkException(ex);
                        }
                    }
                    taskExecutor.shutdown();
                }
            };

            Future<?> generatorResult = taskGenerator.submit(generator);

            int counter = 0;
            while (!generatorResult.isDone()
                    || !futures.isEmpty()) {
                Future<T> containableFuture = futures.take();
                T containable = containableFuture.get();
                // due CascadeType.PERSIST -> also saves Compound...
                containable = save(containable);
                importResult.addImportedEntity(containable.getId());
                counter++;
                // Set hibernate.jdbc.batch_size to 100 too
                // massive performance increase with large inserts
                if (counter >= 100) {
                    counter = 0;
                    entityManager.flush();
                    entityManager.clear();
                    logger.debug("Batch Inserting Containables: entityManager.flush() and entityManager.clear() called.");
                }
            }
            // if not called all exceptions thrown in task generator
            // will be "lost" and method would complete successfully.
            generatorResult.get();
            logger.exit(importResult);
            return importResult;
        } catch (InterruptedException ex) {
            Thread.currentThread().interrupt();
            throw new MoleculeDatabaseFrameworkException(ex);
        } catch (ExecutionException ex) {
            throw new MoleculeDatabaseFrameworkException(ex);
        } finally {
            taskExecutor.shutdownNow();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void exportSDF(Collection<Long> containableIds, Writer writer,
            Set<String> containableProperties, Set<String> compoundProperties) {

        logger.entry(containableIds, writer, containableProperties,
                compoundProperties);
        Preconditions.checkNotNull(writer);
        Preconditions.checkNotNull(containableIds);

        SdfContainableExporter<T> exporter = new SdfContainableExporter<>(this);

        exporter.exportContainablesToSdf(containableIds, writer,
                containableProperties, compoundProperties);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void exportSDF(Collection<Long> containableIds, Writer writer,
            Set<String> containableProperties, Set<String> compoundProperties,
            String dateFormat) {

        logger.entry(containableIds, writer, containableProperties,
                compoundProperties, dateFormat);
        Preconditions.checkNotNull(writer);
        Preconditions.checkNotNull(containableIds);
        Preconditions.checkNotNull(dateFormat);

        SdfContainableExporter<T> exporter = new SdfContainableExporter<>(this);

        exporter.exportContainablesToSdf(containableIds, writer,
                containableProperties, compoundProperties, dateFormat);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final void delete(T containable) {
        logger.entry(containable);
        Preconditions.checkNotNull(containable);
        getRepository().delete(containable);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final void deleteById(Long containableId) {
        logger.entry(containableId);
        Preconditions.checkNotNull(containableId);
        getRepository().delete(containableId);
        logger.exit();
    }

    protected final List<T> toList(Iterable<T> iterable, boolean loadCompositions) {
        if (iterable == null) {
            return null;
        }
        List<T> containers = new ArrayList<>();
        for (T container : iterable) {
            containers.add(container);
            loadCompositions(container, loadCompositions);
        }
        return containers;
    }

    protected final void loadCompositions(Iterable<T> page, boolean loadCompositions) {
        if (page != null && loadCompositions) {
            for (T containable : page) {
                loadCompositions(containable, loadCompositions);
            }
        }
    }

    protected final void loadCompositions(T containable, boolean loadCompositions) {
        if (containable != null) {
            if (loadCompositions) {
                for (ChemicalCompoundComposition composition :
                        (Set<ChemicalCompoundComposition>) containable
                        .getChemicalCompound().getCompositions()) {
                    composition.getChemicalStructure().getId();
                }
            }
        }
    }
}
