/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.google.common.base.Preconditions;
import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import java.util.HashMap;
import lombok.Getter;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.User;
import org.bitbucket.kienerj.moleculedatabaseframework.repository.UserRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * Default implementation of <code>UserService</code>.
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Service("userService")
@Transactional(readOnly = true)
public class UserServiceImpl<T extends User> implements UserService<T> {

    @Autowired
    private UserRepository<T> userRepository;
    @Getter
    private Class<T> entityClass;

    public UserServiceImpl(){
        this.entityClass = (Class<T>)User.class;
    }

    public UserServiceImpl(Class<T> userClass){
        this.entityClass = userClass;
    }

    @Override
    public final UserDetails loadUserByUsername(String username) throws UsernameNotFoundException {
        UserDetails user = this.userRepository.findByUsername(username);
        // required by interface contract
        if (user == null
                || user.getAuthorities() == null
                || user.getAuthorities().isEmpty()){
            throw new UsernameNotFoundException("No user with username " + username + " was found.");
        }
        return user;
    }

    @Override
    public final T getById(Long id) {
        return this.userRepository.findOne(id);
    }

    @Override
    public final T findByUsername(String username){
        return this.userRepository.findByUsername(username);
    }

    @Override
    public final T findOne(Predicate predicate){
         return this.userRepository.findOne(predicate);
    }

    @Override
    public final Iterable<T> findAll(Predicate predicate){
         return this.userRepository.findAll(predicate);
    }

    @Override
    public final Iterable<T> findAll(Predicate predicate, OrderSpecifier<?>... orders){
         return this.userRepository.findAll(predicate, orders);
    }

    @Override
    public final Page<T> findAll(Predicate predicate, Pageable pageable){
         return this.userRepository.findAll(predicate, pageable);
    }

    @Transactional(readOnly = false)
    @Override
    public final T save(T user){
         return this.userRepository.save(user);
    }

    @Transactional(readOnly = false)
    @Override
    public final void delete(T user){
        this.userRepository.delete(user);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public UniquenesscheckResult checkUniqueness(T user) {
        Preconditions.checkNotNull(user);
        Preconditions.checkNotNull(user.getUsername());
        T result = this.findByUsername(user.getUsername());

        HashMap<String, Object> fields = new HashMap<>();
        boolean isUnique = true;

        if (result != null){
            fields.put("userName", user.getUsername());
            isUnique = false;
        }
        return new UniquenesscheckResult(isUnique, fields);
    }
}
